#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Firebird log microservice
# FILE:           saturnin/micro/fblog/api.py
# DESCRIPTION:    API for FBLOG microservice
# CREATED:        13.9.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - API for firebird-log microservice

The FBLOG microservice can perform the following operations with Firebird Server log,
depending on the configuration:

LOG_FROM_SERVER - Pipe OUTPUT: Text log obtained from Firebird server via FB service manager
PARSE_LOG       - Pipe INPUT: Text log, Pipe OUTPUT: Parsed log
FILTER_PARSED   - Pipe INPUT: Parsed log, Pipe OUTPUT: Filtered parsed log
PRINT_PARSED    - Pipe INPUT: Parsed log, Pipe OUTPUT: Text log
"""

import uuid
from functools import partial, reduce
from saturnin.sdk import VENDOR_UID
from saturnin.sdk.types import Enum, AgentDescriptor, ServiceDescriptor, \
     ExecutionMode, ServiceType, ServiceFacilities, SocketMode
from saturnin.sdk.config import create_config, MicroserviceConfig, StrOption, EnumOption, \
     IntOption, BoolOption, ZMQAddressOption, MIMEOption, Option

# OID: iso.org.dod.internet.private.enterprise.firebird.butler.microservice.firebird-log
SERVICE_OID: str = '1.3.6.1.4.1.53446.1.6.2'
SERVICE_UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, SERVICE_OID)
SERVICE_VERSION: str = '0.1'

# Enums

class FbLogOperation(Enum):
    "Operations implemented by FBLOG microservice"
    LOG_FROM_SERVER = Enum.auto()
    PARSE_LOG = Enum.auto()
    FILTER_PARSED = Enum.auto()
    PRINT_PARSED = Enum.auto()

# Configuration

class FbLogConfig(MicroserviceConfig):
    "TextIO service configuration"
    def __init__(self, name: str, description: str):
        super().__init__(name, description)
        self.operation: EnumOption = \
            self.add_option(EnumOption('operation', FbLogOperation,
                                       "Operation to perform", required=True))
        self.stop_on_close: BoolOption = \
            self.add_option(BoolOption('stop_on_close', "Stop service when pipe is closed",
                                       default=True))
        self.input_pipe: StrOption = \
            self.add_option(StrOption('input_pipe', "Input Data Pipe Identification"))
        self.input_address: ZMQAddressOption = \
            self.add_option(ZMQAddressOption('input_address', "Input Data Pipe endpoint address"))
        self.input_mode: EnumOption = \
            self.add_option(EnumOption('input_mode', SocketMode, "Input Data Pipe Mode",
                                       default=SocketMode.CONNECT))
        self.input_format: MIMEOption = \
            self.add_option(MIMEOption('input_format', "Input Pipe data format specification"))
        self.input_batch_size: IntOption = \
            self.add_option(IntOption('input_batch_size', "Input data batch size",
                                      default=50))
        self.output_pipe: StrOption = \
            self.add_option(StrOption('output_pipe', "Output Data Pipe Identification",
                                      required=True))
        self.output_address: ZMQAddressOption = \
            self.add_option(ZMQAddressOption('output_address', "Output Data Pipe endpoint address",
                                             required=True))
        self.output_mode: EnumOption = \
            self.add_option(EnumOption('output_mode', SocketMode, "Output Data Pipe Mode",
                                       required=True, default=SocketMode.CONNECT))
        self.output_format: MIMEOption = \
            self.add_option(MIMEOption('output_format', "Output Pipe data format specification"))
        self.output_batch_size: IntOption = \
            self.add_option(IntOption('output_batch_size', "Output data batch size",
                                      required=True, default=50))
        self.print_format: StrOption = \
            self.add_option(StrOption('print_format', "Format for 'print_parsed' operation",
                                      default='%(origin)s %(timestamp)s [%(level)s] [%(code)s] : %(message)s'))
        self.user: StrOption = \
            self.add_option(StrOption('user', "Firebird user name", default='SYSDBA'))
        self.password: StrOption = \
            self.add_option(StrOption('password', "Firebird user password", default='masterkey'))
        self.filter_expr: StrOption = \
            self.add_option(StrOption('filter_expr', "Filter expression"))
    def __fail(self, msg: str) -> None:
        raise ValueError(msg % self.operation.value.name)
    def __ensure_input(self) -> None:
        if not reduce(lambda res, opt: res and opt is not None,
                      [self.input_pipe, self.input_mode, self.input_address,
                       self.input_batch_size], True):
            self.__fail("Input pipe specification required for operation %s")
    def __ensure_format(self, opt: MIMEOption, fmt: str, direction: str):
        if opt.value and opt.mime_type != fmt:
            raise ValueError("Only '%s' %s format is supported for operation %s"
                             % (fmt, direction, self.operation.value.name))
    def __ensure_required(self, opt: Option) -> None:
        opt.required = True
        opt.validate()
    def validate(self) -> None:
        "Extended validation"
        super().validate()
        if self.operation.value == FbLogOperation.LOG_FROM_SERVER:
            if self.input_pipe.value is not None:
                self.__fail("Input pipe specification not allowed for operation %s")
            self.__ensure_format(self.output_format, 'text/plain', 'output')
            self.__ensure_required(self.user)
            self.__ensure_required(self.password)
        elif self.operation.value == FbLogOperation.PARSE_LOG:
            self.__ensure_input()
            self.__ensure_format(self.input_format, 'text/plain', 'input')
            self.__ensure_format(self.output_format, 'application/x.fb.proto', 'output')
        elif self.operation.value == FbLogOperation.FILTER_PARSED:
            self.__ensure_input()
            self.__ensure_format(self.input_format, 'application/x.fb.proto', 'input')
            self.__ensure_format(self.output_format, 'application/x.fb.proto', 'output')
            self.__ensure_required(self.filter_expr)
        elif self.operation.value == FbLogOperation.PRINT_PARSED:
            self.__ensure_input()
            self.__ensure_format(self.input_format, 'application/x.fb.proto', 'input')
            self.__ensure_format(self.output_format, 'text/plain', 'output')
            self.__ensure_required(self.print_format)

# Service description

SERVICE_AGENT: AgentDescriptor = \
    AgentDescriptor(uid=SERVICE_UID,
                    name="firebird-log",
                    version=SERVICE_VERSION,
                    vendor_uid=VENDOR_UID,
                    classification="firebird/server-log")

SERVICE_DESCRIPTION: ServiceDescriptor = \
    ServiceDescriptor(agent=SERVICE_AGENT,
                      api=[],
                      dependencies=[],
                      execution_mode=ExecutionMode.THREAD,
                      service_type=ServiceType.DATA_FILTER,
                      facilities=(ServiceFacilities.INPUT_CLIENT | ServiceFacilities.INPUT_SERVER |
                                  ServiceFacilities.OUTPUT_CLIENT | ServiceFacilities.OUTPUT_SERVER),
                      description="Microservice for operations with Firebird Server log",
                      implementation='saturnin.micro.fblog.service:FbLogServiceImpl',
                      container='saturnin.sdk.classic:SimpleService',
                      config=partial(create_config, FbLogConfig,
                                     '%s_service' % SERVICE_AGENT.name,
                                     "FBLOG microservice"),
                      client=None,
                      tests=None)



#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Runtime node service
# FILE:           saturnin/service/node/api.py
# DESCRIPTION:    API for Saturnin runtime node service
# CREATED:        12.5.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - Runtime node service API

NODE Service manages Saturnin runtime node. It provides environment for execution and
management of other Saturnin services.

Supported requests:

    :INSTALLED_SERVICES:  REPLY with list of installed services available for execution on the node.
    :RUNNING_SERVICES:    REPLY with list of services actually running on the node.
    :INTERFACE_PROVIDERS: REPLY with list of installed services that provide specified interface.
    :START_SERVICE:       Start service on node.
    :STOP_SERVICE:        Stop service running on node.
    :SHUTDOWN:            Shuts down the NODE service.
"""

import uuid
from functools import partial
from saturnin.sdk import VENDOR_UID
from saturnin.sdk.types import Enum, ExecutionMode, ServiceType, ServiceFacilities, \
     AgentDescriptor, InterfaceDescriptor, ServiceDescriptor
from saturnin.sdk.config import ServiceConfig, create_config


# OID: iso.org.dod.internet.private.enterprise.firebird.butler.service.node
SERVICE_OID: str = '1.3.6.1.4.1.53446.1.1.1'
SERVICE_UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, SERVICE_OID)
SERVICE_VERSION: str = '0.1'

# firebird.butler.service.node.interface
NODE_INTERFACE_OID: str = '1.3.6.1.4.1.53446.1.1.1.0'
NODE_INTERFACE_UID: uuid.UUID = uuid.uuid5(uuid.NAMESPACE_OID, NODE_INTERFACE_OID)

# Enums (Request and Error Codes)

class NodeRequest(Enum):
    "Saturnin Node Service Request Code"
    INSTALLED_SERVICES = 1
    RUNNING_SERVICES = 2
    INTERFACE_PROVIDERS = 3
    START_SERVICE = 4
    STOP_SERVICE = 5
    SHUTDOWN = 6

class NodeError(Enum):
    "Saturnin Node Service Error Code"
    ALREADY_RUNNING = 1
    START_FAILED = 2
    TERMINATION_FAILED = 3

# Service description

SERVICE_AGENT: AgentDescriptor = \
    AgentDescriptor(uid=SERVICE_UID,
                    name="saturnin-node",
                    version=SERVICE_VERSION,
                    vendor_uid=VENDOR_UID,
                    classification="system/runtime")

SERVICE_INTERFACE: InterfaceDescriptor = \
    InterfaceDescriptor(uid=NODE_INTERFACE_UID,
                        name="Saturnin Node service API",
                        revision=1, number=1,
                        requests=NodeRequest)

SERVICE_API = [SERVICE_INTERFACE]

SERVICE_DESCRIPTION: ServiceDescriptor = \
    ServiceDescriptor(agent=SERVICE_AGENT,
                      api=SERVICE_API,
                      dependencies=[],
                      execution_mode=ExecutionMode.ANY,
                      service_type=ServiceType.CONTROL,
                      facilities=ServiceFacilities.FBSP_SOCKET,
                      description="Saturnin runtime node service",
                      implementation='saturnin.service.node.service:SaturninNodeServiceImpl',
                      container='saturnin.sdk.classic:SimpleService',
                      config=partial(create_config, ServiceConfig,
                                     '%s_service' % SERVICE_AGENT.name,
                                     "NODE service."),
                      client='saturnin.service.node.client:SaturninNodeClient',
                      tests='saturnin.service.node.test:TestRunner')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: saturnin/micro/fblog/fblog.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='saturnin/micro/fblog/fblog.proto',
  package='saturnin.protobuf',
  syntax='proto3',
  serialized_pb=_b('\n saturnin/micro/fblog/fblog.proto\x12\x11saturnin.protobuf\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xa8\x01\n\x10\x46irebirdLogEntry\x12\x0e\n\x06origin\x18\x01 \x01(\t\x12-\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05level\x18\x03 \x01(\r\x12\x0c\n\x04\x63ode\x18\x04 \x01(\r\x12\x0f\n\x07message\x18\x05 \x01(\t\x12\'\n\x06params\x18\x06 \x01(\x0b\x32\x17.google.protobuf.Structb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])




_FIREBIRDLOGENTRY = _descriptor.Descriptor(
  name='FirebirdLogEntry',
  full_name='saturnin.protobuf.FirebirdLogEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='origin', full_name='saturnin.protobuf.FirebirdLogEntry.origin', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='saturnin.protobuf.FirebirdLogEntry.timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level', full_name='saturnin.protobuf.FirebirdLogEntry.level', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code', full_name='saturnin.protobuf.FirebirdLogEntry.code', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='saturnin.protobuf.FirebirdLogEntry.message', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='params', full_name='saturnin.protobuf.FirebirdLogEntry.params', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=119,
  serialized_end=287,
)

_FIREBIRDLOGENTRY.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FIREBIRDLOGENTRY.fields_by_name['params'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['FirebirdLogEntry'] = _FIREBIRDLOGENTRY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FirebirdLogEntry = _reflection.GeneratedProtocolMessageType('FirebirdLogEntry', (_message.Message,), dict(
  DESCRIPTOR = _FIREBIRDLOGENTRY,
  __module__ = 'saturnin.micro.fblog.fblog_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.FirebirdLogEntry)
  ))
_sym_db.RegisterMessage(FirebirdLogEntry)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: saturnin/service/node/node.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from firebird.butler import fbsd_pb2 as firebird_dot_butler_dot_fbsd__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='saturnin/service/node/node.proto',
  package='saturnin.protobuf',
  syntax='proto3',
  serialized_pb=_b('\n saturnin/service/node/node.proto\x12\x11saturnin.protobuf\x1a\x19google/protobuf/any.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1a\x66irebird/butler/fbsd.proto\"L\n\nDependency\x12\x31\n\x04type\x18\x01 \x01(\x0e\x32#.firebird.butler.DependencyTypeEnum\x12\x0b\n\x03uid\x18\x02 \x01(\x0c\")\n\x0bRequestCode\x12\x0c\n\x04\x63ode\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x84\x01\n\x13InterfaceDescriptor\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x10\n\x08revision\x18\x03 \x01(\r\x12\x0e\n\x06number\x18\x04 \x01(\r\x12\x30\n\x08requests\x18\x05 \x03(\x0b\x32\x1e.saturnin.protobuf.RequestCode\"\x9c\x03\n\x10InstalledService\x12\x33\n\x05\x61gent\x18\x01 \x01(\x0b\x32$.firebird.butler.AgentIdentification\x12\x33\n\x03\x61pi\x18\x02 \x03(\x0b\x32&.saturnin.protobuf.InterfaceDescriptor\x12\x33\n\x0c\x64\x65pendencies\x18\x03 \x03(\x0b\x32\x1d.saturnin.protobuf.Dependency\x12\x38\n\x0cservice_type\x18\x04 \x01(\x0e\x32\".saturnin.protobuf.ServiceTypeEnum\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x32\n\x04mode\x18\x06 \x01(\x0e\x32$.saturnin.protobuf.ExecutionModeEnum\x12<\n\nfacilities\x18\x07 \x03(\x0e\x32(.saturnin.protobuf.ServiceFacilitiesEnum\x12(\n\nsupplement\x18\x08 \x03(\x0b\x32\x14.google.protobuf.Any\"O\n\x16ReplyInstalledServices\x12\x35\n\x08services\x18\x01 \x03(\x0b\x32#.saturnin.protobuf.InstalledService\"\x9f\x03\n\x0eRunningService\x12\x31\n\x04peer\x18\x01 \x01(\x0b\x32#.firebird.butler.PeerIdentification\x12\x33\n\x05\x61gent\x18\x02 \x01(\x0b\x32$.firebird.butler.AgentIdentification\x12\x38\n\x0cservice_type\x18\x03 \x01(\x0e\x32\".saturnin.protobuf.ServiceTypeEnum\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x32\n\x04mode\x18\x05 \x01(\x0e\x32$.saturnin.protobuf.ExecutionModeEnum\x12<\n\nfacilities\x18\x06 \x03(\x0e\x32(.saturnin.protobuf.ServiceFacilitiesEnum\x12\'\n\x06\x63onfig\x18\x07 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x11\n\tendpoints\x18\x08 \x03(\t\x12(\n\nsupplement\x18\t \x03(\x0b\x32\x14.google.protobuf.Any\"K\n\x14ReplyRunningServices\x12\x33\n\x08services\x18\x01 \x03(\x0b\x32!.saturnin.protobuf.RunningService\"2\n\x19RequestInterfaceProviders\x12\x15\n\rinterface_uid\x18\x01 \x01(\x0c\"-\n\x17ReplyInterfaceProviders\x12\x12\n\nagent_uids\x18\x01 \x03(\x0c\"\x83\x01\n\x13RequestStartService\x12\x11\n\tagent_uid\x18\x01 \x01(\x0c\x12\'\n\x06\x63onfig\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x0f\n\x07timeout\x18\x04 \x01(\r\x12\x11\n\tsingleton\x18\x05 \x01(\x08\"G\n\x11ReplyStartService\x12\x32\n\x07service\x18\x01 \x01(\x0b\x32!.saturnin.protobuf.RunningService\"G\n\x12RequestStopService\x12\x10\n\x08peer_uid\x18\x01 \x01(\x0c\x12\x0f\n\x07timeout\x18\x02 \x01(\r\x12\x0e\n\x06\x66orced\x18\x03 \x01(\x08\">\n\x10ReplyStopService\x12*\n\x06result\x18\x01 \x01(\x0e\x32\x1a.firebird.butler.StateEnum*S\n\x11\x45xecutionModeEnum\x12\x11\n\rEXEC_MODE_ANY\x10\x00\x12\x14\n\x10\x45XEC_MODE_THREAD\x10\x01\x12\x15\n\x11\x45XEC_MODE_PROCESS\x10\x02*\xbf\x01\n\x0fServiceTypeEnum\x12\x14\n\x10SVC_TYPE_UNKNOWN\x10\x00\x12\x1a\n\x16SVC_TYPE_DATA_PROVIDER\x10\x01\x12\x18\n\x14SVC_TYPE_DATA_FILTER\x10\x02\x12\x1a\n\x16SVC_TYPE_DATA_CONSUMER\x10\x03\x12\x17\n\x13SVC_TYPE_PROCESSING\x10\x04\x12\x15\n\x11SVC_TYPE_EXECUTOR\x10\x05\x12\x14\n\x10SVC_TYPE_CONTROL\x10\x06*\xca\x01\n\x15ServiceFacilitiesEnum\x12\x15\n\x11SVC_FACILITY_NONE\x10\x00\x12\x1c\n\x18SVC_FACILITY_FBSP_SOCKET\x10\x01\x12\x1d\n\x19SVC_FACILITY_INPUT_SERVER\x10\x02\x12\x1d\n\x19SVC_FACILITY_INPUT_CLIENT\x10\x03\x12\x1e\n\x1aSVC_FACILITY_OUTPUT_SERVER\x10\x04\x12\x1e\n\x1aSVC_FACILITY_OUTPUT_CLIENT\x10\x05\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,firebird_dot_butler_dot_fbsd__pb2.DESCRIPTOR,])

_EXECUTIONMODEENUM = _descriptor.EnumDescriptor(
  name='ExecutionModeEnum',
  full_name='saturnin.protobuf.ExecutionModeEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXEC_MODE_ANY', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXEC_MODE_THREAD', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXEC_MODE_PROCESS', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1830,
  serialized_end=1913,
)
_sym_db.RegisterEnumDescriptor(_EXECUTIONMODEENUM)

ExecutionModeEnum = enum_type_wrapper.EnumTypeWrapper(_EXECUTIONMODEENUM)
_SERVICETYPEENUM = _descriptor.EnumDescriptor(
  name='ServiceTypeEnum',
  full_name='saturnin.protobuf.ServiceTypeEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_UNKNOWN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_DATA_PROVIDER', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_DATA_FILTER', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_DATA_CONSUMER', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_PROCESSING', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_EXECUTOR', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_TYPE_CONTROL', index=6, number=6,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1916,
  serialized_end=2107,
)
_sym_db.RegisterEnumDescriptor(_SERVICETYPEENUM)

ServiceTypeEnum = enum_type_wrapper.EnumTypeWrapper(_SERVICETYPEENUM)
_SERVICEFACILITIESENUM = _descriptor.EnumDescriptor(
  name='ServiceFacilitiesEnum',
  full_name='saturnin.protobuf.ServiceFacilitiesEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_NONE', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_FBSP_SOCKET', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_INPUT_SERVER', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_INPUT_CLIENT', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_OUTPUT_SERVER', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SVC_FACILITY_OUTPUT_CLIENT', index=5, number=5,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=2110,
  serialized_end=2312,
)
_sym_db.RegisterEnumDescriptor(_SERVICEFACILITIESENUM)

ServiceFacilitiesEnum = enum_type_wrapper.EnumTypeWrapper(_SERVICEFACILITIESENUM)
EXEC_MODE_ANY = 0
EXEC_MODE_THREAD = 1
EXEC_MODE_PROCESS = 2
SVC_TYPE_UNKNOWN = 0
SVC_TYPE_DATA_PROVIDER = 1
SVC_TYPE_DATA_FILTER = 2
SVC_TYPE_DATA_CONSUMER = 3
SVC_TYPE_PROCESSING = 4
SVC_TYPE_EXECUTOR = 5
SVC_TYPE_CONTROL = 6
SVC_FACILITY_NONE = 0
SVC_FACILITY_FBSP_SOCKET = 1
SVC_FACILITY_INPUT_SERVER = 2
SVC_FACILITY_INPUT_CLIENT = 3
SVC_FACILITY_OUTPUT_SERVER = 4
SVC_FACILITY_OUTPUT_CLIENT = 5



_DEPENDENCY = _descriptor.Descriptor(
  name='Dependency',
  full_name='saturnin.protobuf.Dependency',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='saturnin.protobuf.Dependency.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uid', full_name='saturnin.protobuf.Dependency.uid', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=140,
  serialized_end=216,
)


_REQUESTCODE = _descriptor.Descriptor(
  name='RequestCode',
  full_name='saturnin.protobuf.RequestCode',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='saturnin.protobuf.RequestCode.code', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='saturnin.protobuf.RequestCode.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=218,
  serialized_end=259,
)


_INTERFACEDESCRIPTOR = _descriptor.Descriptor(
  name='InterfaceDescriptor',
  full_name='saturnin.protobuf.InterfaceDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uid', full_name='saturnin.protobuf.InterfaceDescriptor.uid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='saturnin.protobuf.InterfaceDescriptor.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='revision', full_name='saturnin.protobuf.InterfaceDescriptor.revision', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='number', full_name='saturnin.protobuf.InterfaceDescriptor.number', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requests', full_name='saturnin.protobuf.InterfaceDescriptor.requests', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=262,
  serialized_end=394,
)


_INSTALLEDSERVICE = _descriptor.Descriptor(
  name='InstalledService',
  full_name='saturnin.protobuf.InstalledService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent', full_name='saturnin.protobuf.InstalledService.agent', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='api', full_name='saturnin.protobuf.InstalledService.api', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dependencies', full_name='saturnin.protobuf.InstalledService.dependencies', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='service_type', full_name='saturnin.protobuf.InstalledService.service_type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='saturnin.protobuf.InstalledService.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='saturnin.protobuf.InstalledService.mode', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='facilities', full_name='saturnin.protobuf.InstalledService.facilities', index=6,
      number=7, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supplement', full_name='saturnin.protobuf.InstalledService.supplement', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=397,
  serialized_end=809,
)


_REPLYINSTALLEDSERVICES = _descriptor.Descriptor(
  name='ReplyInstalledServices',
  full_name='saturnin.protobuf.ReplyInstalledServices',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='services', full_name='saturnin.protobuf.ReplyInstalledServices.services', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=811,
  serialized_end=890,
)


_RUNNINGSERVICE = _descriptor.Descriptor(
  name='RunningService',
  full_name='saturnin.protobuf.RunningService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer', full_name='saturnin.protobuf.RunningService.peer', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='agent', full_name='saturnin.protobuf.RunningService.agent', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='service_type', full_name='saturnin.protobuf.RunningService.service_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='saturnin.protobuf.RunningService.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mode', full_name='saturnin.protobuf.RunningService.mode', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='facilities', full_name='saturnin.protobuf.RunningService.facilities', index=5,
      number=6, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='config', full_name='saturnin.protobuf.RunningService.config', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='endpoints', full_name='saturnin.protobuf.RunningService.endpoints', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='supplement', full_name='saturnin.protobuf.RunningService.supplement', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=893,
  serialized_end=1308,
)


_REPLYRUNNINGSERVICES = _descriptor.Descriptor(
  name='ReplyRunningServices',
  full_name='saturnin.protobuf.ReplyRunningServices',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='services', full_name='saturnin.protobuf.ReplyRunningServices.services', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1310,
  serialized_end=1385,
)


_REQUESTINTERFACEPROVIDERS = _descriptor.Descriptor(
  name='RequestInterfaceProviders',
  full_name='saturnin.protobuf.RequestInterfaceProviders',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='interface_uid', full_name='saturnin.protobuf.RequestInterfaceProviders.interface_uid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1387,
  serialized_end=1437,
)


_REPLYINTERFACEPROVIDERS = _descriptor.Descriptor(
  name='ReplyInterfaceProviders',
  full_name='saturnin.protobuf.ReplyInterfaceProviders',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_uids', full_name='saturnin.protobuf.ReplyInterfaceProviders.agent_uids', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1439,
  serialized_end=1484,
)


_REQUESTSTARTSERVICE = _descriptor.Descriptor(
  name='RequestStartService',
  full_name='saturnin.protobuf.RequestStartService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='agent_uid', full_name='saturnin.protobuf.RequestStartService.agent_uid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='config', full_name='saturnin.protobuf.RequestStartService.config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='saturnin.protobuf.RequestStartService.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timeout', full_name='saturnin.protobuf.RequestStartService.timeout', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='singleton', full_name='saturnin.protobuf.RequestStartService.singleton', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1487,
  serialized_end=1618,
)


_REPLYSTARTSERVICE = _descriptor.Descriptor(
  name='ReplyStartService',
  full_name='saturnin.protobuf.ReplyStartService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='saturnin.protobuf.ReplyStartService.service', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1620,
  serialized_end=1691,
)


_REQUESTSTOPSERVICE = _descriptor.Descriptor(
  name='RequestStopService',
  full_name='saturnin.protobuf.RequestStopService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer_uid', full_name='saturnin.protobuf.RequestStopService.peer_uid', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timeout', full_name='saturnin.protobuf.RequestStopService.timeout', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='forced', full_name='saturnin.protobuf.RequestStopService.forced', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1693,
  serialized_end=1764,
)


_REPLYSTOPSERVICE = _descriptor.Descriptor(
  name='ReplyStopService',
  full_name='saturnin.protobuf.ReplyStopService',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='saturnin.protobuf.ReplyStopService.result', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1766,
  serialized_end=1828,
)

_DEPENDENCY.fields_by_name['type'].enum_type = firebird_dot_butler_dot_fbsd__pb2._DEPENDENCYTYPEENUM
_INTERFACEDESCRIPTOR.fields_by_name['requests'].message_type = _REQUESTCODE
_INSTALLEDSERVICE.fields_by_name['agent'].message_type = firebird_dot_butler_dot_fbsd__pb2._AGENTIDENTIFICATION
_INSTALLEDSERVICE.fields_by_name['api'].message_type = _INTERFACEDESCRIPTOR
_INSTALLEDSERVICE.fields_by_name['dependencies'].message_type = _DEPENDENCY
_INSTALLEDSERVICE.fields_by_name['service_type'].enum_type = _SERVICETYPEENUM
_INSTALLEDSERVICE.fields_by_name['mode'].enum_type = _EXECUTIONMODEENUM
_INSTALLEDSERVICE.fields_by_name['facilities'].enum_type = _SERVICEFACILITIESENUM
_INSTALLEDSERVICE.fields_by_name['supplement'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_REPLYINSTALLEDSERVICES.fields_by_name['services'].message_type = _INSTALLEDSERVICE
_RUNNINGSERVICE.fields_by_name['peer'].message_type = firebird_dot_butler_dot_fbsd__pb2._PEERIDENTIFICATION
_RUNNINGSERVICE.fields_by_name['agent'].message_type = firebird_dot_butler_dot_fbsd__pb2._AGENTIDENTIFICATION
_RUNNINGSERVICE.fields_by_name['service_type'].enum_type = _SERVICETYPEENUM
_RUNNINGSERVICE.fields_by_name['mode'].enum_type = _EXECUTIONMODEENUM
_RUNNINGSERVICE.fields_by_name['facilities'].enum_type = _SERVICEFACILITIESENUM
_RUNNINGSERVICE.fields_by_name['config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_RUNNINGSERVICE.fields_by_name['supplement'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_REPLYRUNNINGSERVICES.fields_by_name['services'].message_type = _RUNNINGSERVICE
_REQUESTSTARTSERVICE.fields_by_name['config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_REPLYSTARTSERVICE.fields_by_name['service'].message_type = _RUNNINGSERVICE
_REPLYSTOPSERVICE.fields_by_name['result'].enum_type = firebird_dot_butler_dot_fbsd__pb2._STATEENUM
DESCRIPTOR.message_types_by_name['Dependency'] = _DEPENDENCY
DESCRIPTOR.message_types_by_name['RequestCode'] = _REQUESTCODE
DESCRIPTOR.message_types_by_name['InterfaceDescriptor'] = _INTERFACEDESCRIPTOR
DESCRIPTOR.message_types_by_name['InstalledService'] = _INSTALLEDSERVICE
DESCRIPTOR.message_types_by_name['ReplyInstalledServices'] = _REPLYINSTALLEDSERVICES
DESCRIPTOR.message_types_by_name['RunningService'] = _RUNNINGSERVICE
DESCRIPTOR.message_types_by_name['ReplyRunningServices'] = _REPLYRUNNINGSERVICES
DESCRIPTOR.message_types_by_name['RequestInterfaceProviders'] = _REQUESTINTERFACEPROVIDERS
DESCRIPTOR.message_types_by_name['ReplyInterfaceProviders'] = _REPLYINTERFACEPROVIDERS
DESCRIPTOR.message_types_by_name['RequestStartService'] = _REQUESTSTARTSERVICE
DESCRIPTOR.message_types_by_name['ReplyStartService'] = _REPLYSTARTSERVICE
DESCRIPTOR.message_types_by_name['RequestStopService'] = _REQUESTSTOPSERVICE
DESCRIPTOR.message_types_by_name['ReplyStopService'] = _REPLYSTOPSERVICE
DESCRIPTOR.enum_types_by_name['ExecutionModeEnum'] = _EXECUTIONMODEENUM
DESCRIPTOR.enum_types_by_name['ServiceTypeEnum'] = _SERVICETYPEENUM
DESCRIPTOR.enum_types_by_name['ServiceFacilitiesEnum'] = _SERVICEFACILITIESENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dependency = _reflection.GeneratedProtocolMessageType('Dependency', (_message.Message,), dict(
  DESCRIPTOR = _DEPENDENCY,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.Dependency)
  ))
_sym_db.RegisterMessage(Dependency)

RequestCode = _reflection.GeneratedProtocolMessageType('RequestCode', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTCODE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.RequestCode)
  ))
_sym_db.RegisterMessage(RequestCode)

InterfaceDescriptor = _reflection.GeneratedProtocolMessageType('InterfaceDescriptor', (_message.Message,), dict(
  DESCRIPTOR = _INTERFACEDESCRIPTOR,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.InterfaceDescriptor)
  ))
_sym_db.RegisterMessage(InterfaceDescriptor)

InstalledService = _reflection.GeneratedProtocolMessageType('InstalledService', (_message.Message,), dict(
  DESCRIPTOR = _INSTALLEDSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.InstalledService)
  ))
_sym_db.RegisterMessage(InstalledService)

ReplyInstalledServices = _reflection.GeneratedProtocolMessageType('ReplyInstalledServices', (_message.Message,), dict(
  DESCRIPTOR = _REPLYINSTALLEDSERVICES,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.ReplyInstalledServices)
  ))
_sym_db.RegisterMessage(ReplyInstalledServices)

RunningService = _reflection.GeneratedProtocolMessageType('RunningService', (_message.Message,), dict(
  DESCRIPTOR = _RUNNINGSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.RunningService)
  ))
_sym_db.RegisterMessage(RunningService)

ReplyRunningServices = _reflection.GeneratedProtocolMessageType('ReplyRunningServices', (_message.Message,), dict(
  DESCRIPTOR = _REPLYRUNNINGSERVICES,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.ReplyRunningServices)
  ))
_sym_db.RegisterMessage(ReplyRunningServices)

RequestInterfaceProviders = _reflection.GeneratedProtocolMessageType('RequestInterfaceProviders', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTINTERFACEPROVIDERS,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.RequestInterfaceProviders)
  ))
_sym_db.RegisterMessage(RequestInterfaceProviders)

ReplyInterfaceProviders = _reflection.GeneratedProtocolMessageType('ReplyInterfaceProviders', (_message.Message,), dict(
  DESCRIPTOR = _REPLYINTERFACEPROVIDERS,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.ReplyInterfaceProviders)
  ))
_sym_db.RegisterMessage(ReplyInterfaceProviders)

RequestStartService = _reflection.GeneratedProtocolMessageType('RequestStartService', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTSTARTSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.RequestStartService)
  ))
_sym_db.RegisterMessage(RequestStartService)

ReplyStartService = _reflection.GeneratedProtocolMessageType('ReplyStartService', (_message.Message,), dict(
  DESCRIPTOR = _REPLYSTARTSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.ReplyStartService)
  ))
_sym_db.RegisterMessage(ReplyStartService)

RequestStopService = _reflection.GeneratedProtocolMessageType('RequestStopService', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTSTOPSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.RequestStopService)
  ))
_sym_db.RegisterMessage(RequestStopService)

ReplyStopService = _reflection.GeneratedProtocolMessageType('ReplyStopService', (_message.Message,), dict(
  DESCRIPTOR = _REPLYSTOPSERVICE,
  __module__ = 'saturnin.service.node.node_pb2'
  # @@protoc_insertion_point(class_scope:saturnin.protobuf.ReplyStopService)
  ))
_sym_db.RegisterMessage(ReplyStopService)


# @@protoc_insertion_point(module_scope)



#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Firebird log microservice
# FILE:           saturnin/micro/fblog/service.py
# DESCRIPTION:    FBLOG microservice (classic version)
# CREATED:        13.9.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - FBLOG microservice

The FBLOG microservice can perform the following operations with Firebird Server log,
depending on the configuration:

LOG_FROM_SERVER - Pipe OUTPUT: Text log obtained from Firebird server via FB service manager
PARSE_LOG       - Pipe INPUT: Text log, Pipe OUTPUT: Parsed log
FILTER_PARSED   - Pipe INPUT: Parsed log, Pipe OUTPUT: Filtered parsed log
PRINT_PARSED    - Pipe INPUT: Parsed log, Pipe OUTPUT: Text log
"""


import logging
import typing as t
from datetime import datetime
from saturnin.sdk.types import Enum, SocketMode, ZMQAddress, ServiceDescriptor, StopError
from saturnin.sdk.config import MIMEOption
from saturnin.sdk.base import DealerChannel
from saturnin.sdk.protocol.fbdp import PipeSocket, BaseFBDPHandler, PipeClientHandler, \
     PipeServerHandler, Session, ErrorCode, MsgType, Message
from saturnin.sdk.service import MicroserviceImpl, BaseService
from .api import FbLogConfig, FbLogOperation
from . import fblog_pb2 as pb

# Logger

log = logging.getLogger(__name__)

LOG_FORMAT = 'application/x.fb.proto;type=saturnin.micro.fblog.FirebirdLogEntry'

END_OF_DATA = object()

# Enums

class PipeState(Enum):
    "Data Pipe state"
    UNKNOWN = 0
    OPEN = 1
    READY = 2
    TRANSMITTING = 3
    CLOSED = 4

# Types

TOnPipeClosed = t.Callable[['DataPipe', Session, Message], None]
TOnAcceptClient = t.Callable[['DataPipe', Session, MIMEOption], int]
TOnServerReady = t.Callable[['DataPipe', Session, int], int]
TOnAcceptData = t.Callable[['DataPipe', Session, bytes], int]
TOnProduceData = t.Callable[['DataPipe', Session], t.Any]

# Functions

def value(value: t.Any, default: t.Any) -> t.Any:
    return default if value is None else value

# Classes

class DataPipe:
    """Data pipe descriptor.

Attributes:
    :state:       Pipe state
    :pipe_id:     Pipe identification
    :mode:        Pipe mode
    :address:     Pipe ZMQ address
    :socket:      Pipe socket
    :data_format: Data format
    :mime_type:   MIME type
    :mime_params: MIME type parameters
    :chn:         Pipe channel

Abstract methods:
    :on_pipe_closed:   General callback called when pipe is closed.
    :on_accept_client: SERVER callback. Validates and processes client connection request.
    :on_server_ready:  CLIENT callback executed when non-zero READY is received from server.
    :on_accept_data:   CONSUMER callback that process DATA from producer.
    :on_produce_data:  PRODUCER callback that returns DATA for consumer.
"""
    def __init__(self, on_close: TOnPipeClosed = None):
        self.state = PipeState.UNKNOWN
        self.chn: DealerChannel = None
        self.hnd: BaseFBDPHandler = None
        self.pipe_id: str = None
        self.mode: SocketMode = None
        self.address: ZMQAddress = None
        self.socket: PipeSocket = None
        self.batch_size: int = 50
        self.data_format: str = ''
        self.mime_type: str = None
        self.mime_params: t.Dict[str, str] = None
        self.active: bool = False
        #
        self.on_pipe_closed: TOnPipeClosed = on_close
        self.on_accept_client: TOnAcceptClient = None
        self.on_server_ready: TOnServerReady = None
        self.on_accept_data: TOnAcceptData = None
        self.on_produce_data: TOnProduceData = None
    def __on_accept_client(self, handler: BaseFBDPHandler, session: Session,
                         data_pipe: str, pipe_stream: PipeSocket, data_format: str) -> int:
        "SERVER callback. Validates and processes client connection request."
        log.debug('%s._on_accept_client(%s,%s,%s) [%s]', self.__class__.__name__, data_pipe,
                  pipe_stream, data_format, self.pipe_id)
        if data_pipe != self.pipe_id:
            raise StopError("Unknown data pipe '%s'" % data_pipe,
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        elif pipe_stream != self.socket:
            raise StopError("'%s' stream not available" % pipe_stream.name,
                            code = ErrorCode.PIPE_ENDPOINT_UNAVAILABLE)
        mime = MIMEOption('data_format', '')
        try:
            mime.set_value(data_format)
        except:
            StopError("Only MIME data formats supported",
                      code = ErrorCode.DATA_FORMAT_NOT_SUPPORTED)
        result = session.batch_size
        if self.on_accept_client:
            result = self.on_accept_client(self, session, mime)
        self.active = True
        return result
    def __on_server_ready(self, handler: PipeClientHandler, session: Session, batch_size: int) -> int:
        self.active = True
        if self.on_server_ready:
            return self.on_server_ready(self, session, batch_size)
        return min(batch_size, session.batch_size)
    def __on_batch_start(self, handler: BaseFBDPHandler, session: Session) -> None:
        while session.transmit > 0:
            try:
                data: t.Any = self.on_produce_data(self, session)
            except StopError:
                handler.send_close(session, ErrorCode.OK)
                return
            except Exception as exc:
                handler.send_close(session, ErrorCode.INVALID_DATA, exc=exc)
                return
            if data is None:
                # Data are not available at this time, schedule for later
                self.chn.manager.defer(self.__on_batch_start, handler, session)
                return
            msg = handler.protocol.create_message_for(MsgType.DATA)
            msg.data_frame = data
            session.transmit -= 1
            handler.send(msg, session)
        if self.mode == SocketMode.BIND and session.transmit == 0:
            try:
                self.hnd.send_ready(session, handler.batch_size)
            except Exception as exc:
                exc.__traceback__ = None
                log.error("Pipe READY send failed [SRV:%s:%s]", session.pipe_stream,
                          session.data_pipe, exc_info=exc)
    def __on_accept_data(self, handler: BaseFBDPHandler, session: Session, data: t.Any) -> int:
        assert self.on_accept_data
        return self.on_accept_data(self, session, data)
    def __on_pipe_closed(self, handler: BaseFBDPHandler, session: Session, msg: Message) -> None:
        log.debug('%s.__on_pipe_closed() [%s]', self.__class__.__name__, self.pipe_id)
        self.active = False
        if self.on_pipe_closed:
            self.on_pipe_closed(self, session, msg)
    def set_channel(self, channel: DealerChannel) -> None:
        "Assign transmission channel for data pipe."
        self.chn = channel
    def set_mode(self, mode: SocketMode) -> None:
        assert self.chn is not None
        self.mode = mode
        if mode == SocketMode.BIND:
            self.hnd = PipeServerHandler(self.batch_size)
            self.hnd.on_accept_client = self.__on_accept_client
        else:
            self.hnd = PipeClientHandler(self.batch_size)
            self.hnd.on_server_ready = self.__on_server_ready
        self.hnd.on_batch_start = self.__on_batch_start
        self.hnd.on_accept_data = self.__on_accept_data
        self.hnd.on_pipe_closed = self.__on_pipe_closed
        self.chn.set_handler(self.hnd)
    def set_format(self, mime: MIMEOption, default: str = None) -> None:
        "Set MIME data format"
        fmt = mime
        if mime.value is None and default:
            fmt = MIMEOption('', '')
            fmt.set_value(default)
        if fmt.value is None:
            raise ValueError("MIME format not specified")
        self.data_format = fmt.value
        self.mime_type = fmt.mime_type
        self.mime_params = dict(fmt.mime_params)
    def close(self) -> None:
        "Close the data pipe"
        self.hnd.close()
    def open(self) -> None:
        "Open the data pipe"
        assert self.chn is not None
        assert self.hnd is not None
        assert self.pipe_id is not None
        assert self.mode is not None
        assert self.address is not None
        assert self.socket is not None
        if self.mode == SocketMode.CONNECT:
            self.hnd.open(self.address, self.pipe_id, self.socket, self.data_format)
        else:
            self.chn.bind(self.address)

class InputPipe(DataPipe):
    "INPUT data pipe"
    def open(self) -> None:
        "Open the data pipe"
        assert self.chn is not None
        assert self.hnd is not None
        assert self.pipe_id is not None
        assert self.mode is not None
        assert self.address is not None
        if self.mode == SocketMode.CONNECT:
            self.socket = PipeSocket.OUTPUT
            self.hnd.open(self.address, self.pipe_id, self.socket, self.data_format)
        else:
            self.socket = PipeSocket.INPUT
            log.info("Pipe BIND [%s:%s]", t.cast(PipeSocket, self.socket).name, self.pipe_id)
            self.chn.bind(self.address)

class OutputPipe(DataPipe):
    "OUTPUT data pipe"
    def open(self) -> None:
        "Open the data pipe"
        assert self.chn is not None
        assert self.hnd is not None
        assert self.pipe_id is not None
        assert self.mode is not None
        assert self.address is not None
        if self.mode == SocketMode.CONNECT:
            self.socket = PipeSocket.INPUT
            self.hnd.open(self.address, self.pipe_id, self.socket, self.data_format)
        else:
            self.socket = PipeSocket.OUTPUT
            log.info("Pipe BIND [%s:%s]", t.cast(PipeSocket, self.socket).name, self.pipe_id)
            self.chn.bind(self.address)

class FbLogServiceImpl(MicroserviceImpl):
    """Implementation of FBLOG microservice."""
    def __init__(self, descriptor: ServiceDescriptor, stop_event: t.Any):
        super().__init__(descriptor, stop_event)
        self.in_pipe: InputPipe = InputPipe(self.on_pipe_closed)
        self.out_pipe: OutputPipe = OutputPipe(self.on_pipe_closed)
        self.in_chn: DealerChannel = None
        self.out_chn: DealerChannel = None
        self.stop_on_close = True
        self.operation: FbLogOperation = None
        self.in_que: t.List = []
        self.send_que: t.List = []
        self.proto: pb.FirebirdLogEntry = pb.FirebirdLogEntry()
        #self.ready = False
        self.post_process = None
        self.print_format: str = None
        self.filter_expr: str = None
        self.print_cnt = 0
        self.svc = None
        #
    def __parse(self) -> None:
        "Parse input queue and store result into send queue."
        origin, timestamp = self.in_que.pop(0)
        self.proto.Clear()
        self.proto.origin = origin
        self.proto.timestamp.FromDatetime(timestamp)
        self.proto.message = ' '.join(self.in_que)
        self.send_que.append(self.proto.SerializeToString())
        self.in_que.clear()
    def on_accept_client(self, pipe: DataPipe, session: Session, fmt: MIMEOption) -> int:
        "Either reject client by StopError, or return batch_size we are ready to transmit."
        if __debug__: log.debug('%s.on_accept_client [%s]', self.__class__.__name__, pipe.pipe_id)
        other: DataPipe = self.out_pipe if pipe is self.in_pipe else self.in_pipe
        result = 0
        if other.active():
            result = min(pipe.batch_size, other.batch_size, session.batch_size)
        return result
    def on_server_ready(self, pipe: DataPipe, session: Session, batch_size: int) -> int:
        "Default callback that returns batch_size if < session.batch_size, else session.batch_size."
        if __debug__: log.debug('%s.on_server_ready [%s]', self.__class__.__name__, pipe.pipe_id)
        # session.ready is None for first READY from server
        if session.ready is None:
            other: DataPipe = self.out_pipe if pipe is self.in_pipe else self.in_pipe
            result = 0
            if other.active():
                result = min(pipe.batch_size, other.batch_size, session.batch_size)
            return result
        return min(batch_size, session.batch_size)
    def accept_print_entry(self, pipe: DataPipe, session: Session, data: bytes) -> t.Optional[int]:
        "Process parsed log entry"
        if __debug__: log.debug('%s.accept_print_entry [%s]', self.__class__.__name__, pipe.pipe_id)
        assert self.print_format
        self.proto.Clear()
        self.print_cnt += 1
        try:
            self.proto.ParseFromString(data)
            self.send_que.append(bytes(self.print_format
                                       % {'lineno': self.print_cnt,
                                          'origin': self.proto.origin,
                                          'timestamp': self.proto.timestamp.ToDatetime(),
                                          'level': self.proto.level,
                                          'code': self.proto.code,
                                          'message': self.proto.message},
                                       self.out_pipe.mime_params.get('charset', 'utf-8'),
                                       self.out_pipe.mime_params.get('errors', 'strict')))
        except:
            return ErrorCode.INVALID_DATA
    def accept_filter_entry(self, pipe: DataPipe, session: Session, data: bytes) -> t.Optional[int]:
        "Process parsed log entry"
        if __debug__: log.debug('%s.accept_filter_entry [%s]', self.__class__.__name__, pipe.pipe_id)
        assert self.filter_expr
        self.proto.Clear()
        try:
            # Filter entry
            self.proto.ParseFromString(data)
            _locals = {'origin': self.proto.origin, 'code': self.proto.code,
                       'timestamp': self.proto.timestamp.ToDatetime(),
                       'level': self.proto.level, 'message': self.proto.message}
            _locals.update(self.proto.params)
            if eval(self.filter_expr, None, _locals):
                self.send_que.append(data)
        except:
            log.exception("FILTER ERROR:")
            return ErrorCode.INVALID_DATA
    def accept_log_line(self, pipe: DataPipe, session: Session, data: bytes) -> t.Optional[int]:
        "Process text log line"
        if __debug__: log.debug('%s.accept_log_line [%s]', self.__class__.__name__, pipe.pipe_id)
        try:
            line = data.decode(encoding=pipe.mime_params.get('charset', 'utf-8'),
                               errors=pipe.mime_params.get('errors', 'strict')).strip()
        except:
            return ErrorCode.INVALID_DATA
        #
        if not line:
            if self.in_que:
                self.__parse()
            return
        if not self.in_que:
            items = line.split()
            try:
                timestamp = datetime.strptime(' '.join(items[len(items)-5:]),
                                              '%a %b %d %H:%M:%S %Y')
            except ValueError:
                return ErrorCode.INVALID_DATA
            origin = ' '.join(items[:len(items)-5])
            self.in_que.append((origin, timestamp))
        else:
            self.in_que.append(line)
    def produce_output(self, pipe: DataPipe, session: Session) -> pb.FirebirdLogEntry:
        "Return entry from queue or None"
        if __debug__: log.debug('%s.produce_output [%s]', self.__class__.__name__, pipe.pipe_id)
        data = None
        if self.operation == FbLogOperation.LOG_FROM_SERVER:
            data = self.svc.readline()
            if data is None:
                data = END_OF_DATA
            else:
                data = bytes(data, self.out_pipe.mime_params.get('charset', 'utf-8'),
                             self.out_pipe.mime_params.get('errors', 'strict'))
        else:
            if self.send_que:
                data = self.send_que.pop(0)
        if data is END_OF_DATA:
            raise StopError("EOF")
        return data
    def on_pipe_closed(self, pipe: DataPipe, session: Session, msg: Message) -> None:
        "General callback that logs info(OK) or error, and closes the input file."
        if __debug__: log.debug('%s.on_pipe_closed [%s]', self.__class__.__name__, pipe.pipe_id)
        # If output was closed, close input as well
        if pipe is self.out_pipe:
            if self.in_pipe.hnd.is_active():
                self.in_pipe.close()
            self.send_que.clear() # clear the out queue, we may wait for another consumer
            self.print_cnt = 0
            self.filter_expr = None
        elif pipe is self.in_pipe:
            # if we still have data in input queue and pipe was closed normally, process them
            if self.in_que and msg.type_data == ErrorCode.OK:
                # this can turn out any way, so ignore errors
                try:
                    assert self.post_process
                    self.post_process()
                except:
                    pass
            self.in_que.clear() # clear the in queue, we may wait for another producer
            # If we have active output, signal end of data
            if self.out_pipe.hnd.is_active():
                self.send_que.append(END_OF_DATA)
        if (not self.in_pipe.hnd.is_active() and
            not self.out_pipe.hnd.is_active() and self.stop_on_close):
            self.stop_event.set()
    def initialize(self, svc: BaseService):
        super().initialize(svc)
        self.in_chn = DealerChannel(b'%s-in' % self.instance_id.hex().encode('ascii'),
                                      sock_opts=self.get('in_sock_opts', None))
        self.out_chn = DealerChannel(b'%s-out' % self.instance_id.hex().encode('ascii'),
                                      sock_opts=self.get('out_sock_opts', None))
        self.in_pipe.set_channel(self.in_chn)
        self.out_pipe.set_channel(self.out_chn)
        self.mngr.add(self.in_chn)
        self.mngr.add(self.out_chn)
    def configure(self, svc: BaseService, config: FbLogConfig) -> None:
        """Service configuration."""
        config.validate() # Fail early
        #
        self.operation = config.operation.value
        self.stop_on_close = config.stop_on_close.value
        #
        self.in_pipe.pipe_id = config.input_pipe.value
        self.in_pipe.address = config.input_address.value
        self.in_pipe.set_mode(config.input_mode.value)
        self.in_pipe.batch_size = config.input_batch_size.value
        #
        self.out_pipe.pipe_id = config.output_pipe.value
        self.out_pipe.address = config.output_address.value
        self.out_pipe.set_mode(config.output_mode.value)
        self.out_pipe.batch_size = config.output_batch_size.value
        # TODO: Validate data format specifications for pipes
        # Prepare selected operation
        if self.operation == FbLogOperation.LOG_FROM_SERVER:
            self.out_pipe.set_format(config.output_format, 'text/plain;charset=utf-8')
            from fdb.services import connect
            self.svc = connect(user=config.user.value, password=config.password.value)
            self.svc.get_log()
            self.out_pipe.on_produce_data = self.produce_output
        elif self.operation == FbLogOperation.PARSE_LOG:
            self.in_pipe.set_format(config.input_format, 'text/plain;charset=utf-8')
            self.out_pipe.set_format(config.output_format, LOG_FORMAT)
            #
            self.in_pipe.on_accept_data = self.accept_log_line
            self.post_process = self.__parse
            self.out_pipe.on_produce_data = self.produce_output
            #
        elif self.operation == FbLogOperation.FILTER_PARSED:
            self.in_pipe.set_format(config.input_format, LOG_FORMAT)
            self.out_pipe.set_format(config.output_format, LOG_FORMAT)
            #
            self.in_pipe.on_accept_data = self.accept_filter_entry
            try:
                self.filter_expr = compile(config.filter_expr.value, '<filter_expr>', 'eval',
                                           optimize=2)
            except Exception as exc:
                raise StopError("Error in filter expression", exc) from None
            self.out_pipe.on_produce_data = self.produce_output
        elif self.operation == FbLogOperation.PRINT_PARSED:
            self.in_pipe.set_format(config.input_format, LOG_FORMAT)
            self.out_pipe.set_format(config.output_format, 'text/plain;charset=utf-8')
            #
            self.in_pipe.on_accept_data = self.accept_print_entry
            self.print_format = config.print_format.value
            if not self.print_format.endswith('\n'):
                self.print_format += '\n'
            self.out_pipe.on_produce_data = self.produce_output
            #
        else:
            raise StopError("Operation '%s' not implemented" % self.operation.name)
        # Open pipes
        for pipe in (self.in_pipe, self.out_pipe):
            if pipe.pipe_id:
                if pipe.mode == SocketMode.BIND:
                    pipe.open()
                else:
                    self.mngr.defer(pipe.open)
        #
    def validate(self) -> None:
        ""
    def finalize(self, svc: BaseService) -> None:
        """Service finalization."""
        for pipe in (self.in_pipe, self.out_pipe):
            pipe.close()
        super().finalize(svc)



#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Runtime node service
# FILE:           saturnin/service/node/client.py
# DESCRIPTION:    NODE service client (classic version)
# CREATED:        12.5.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - NODE service client (classic version)

NODE Service manages Saturnin runtime node. It provides environment for execution and
management of other Saturnin services.
"""

import typing as t
from uuid import UUID
from saturnin.sdk.types import State, InterfaceDescriptor
from saturnin.sdk.config import Config
from saturnin.sdk.protocol.fbsp import Session, MsgType, bb2h, ReplyMessage, ErrorMessage, exception_for
from saturnin.sdk.client import ServiceClient
from . import node_pb2 as pb
from .api import NodeRequest, SERVICE_INTERFACE

class SaturninNodeClient(ServiceClient):
    """Message handler for NODE client."""
    def get_interface(self) -> InterfaceDescriptor:
        return SERVICE_INTERFACE
    def get_handlers(self, api_number: int) -> t.Dict:
        return {(MsgType.REPLY, bb2h(api_number, NodeRequest.INSTALLED_SERVICES)):
                self.handle_installed,
                (MsgType.REPLY, bb2h(api_number, NodeRequest.RUNNING_SERVICES)):
                self.handle_running,
                (MsgType.REPLY, bb2h(api_number, NodeRequest.INTERFACE_PROVIDERS)):
                self.handle_providers,
                (MsgType.REPLY, bb2h(api_number, NodeRequest.START_SERVICE)):
                self.handle_start,
                (MsgType.REPLY, bb2h(api_number, NodeRequest.STOP_SERVICE)):
                self.handle_stop,
                (MsgType.REPLY, bb2h(api_number, NodeRequest.SHUTDOWN)):
                self.handle_shutdown,
                MsgType.DATA: self.raise_protocol_violation,
                MsgType.REPLY: self.raise_protocol_violation,
                MsgType.STATE: self.raise_protocol_violation,
               }
    def handle_error(self, session: Session, msg: ErrorMessage):
        "Handle ERROR message received from Service."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        req.response = None
        if msg.token != session.greeting.token:
            session.request_done(req)
        raise exception_for(msg)
    def handle_installed(self, session: Session, msg: ReplyMessage):
        "INSTALLED_SERVICES reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        dframe = pb.ReplyInstalledServices()
        dframe.ParseFromString(msg.data.pop(0))
        req.response = dframe
        session.request_done(req)
    def handle_running(self, session: Session, msg: ReplyMessage):
        "RUNNING_SERVICES reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        dframe = pb.ReplyRunningServices()
        dframe.ParseFromString(msg.data.pop(0))
        req.response = dframe
        session.request_done(req)
    def handle_providers(self, session: Session, msg: ReplyMessage):
        "INTERFACE_PROVIDERS reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        dframe = pb.ReplyInterfaceProviders()
        dframe.ParseFromString(msg.data.pop(0))
        req.response = [UUID(bytes=uid) for uid in dframe.agent_uids]
        session.request_done(req)
    def handle_start(self, session: Session, msg: ReplyMessage):
        "START_SERVICE reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        dframe = pb.ReplyStartService()
        dframe.ParseFromString(msg.data.pop(0))
        req.response = dframe
        session.request_done(req)
    def handle_stop(self, session: Session, msg: ReplyMessage):
        "STOP_SERVICE reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        dframe = pb.ReplyStopService()
        dframe.ParseFromString(msg.data.pop(0))
        req.response = State(dframe.result)
        session.request_done(req)
    def handle_shutdown(self, session: Session, msg: ReplyMessage):
        "SHUTDOWN reply handler."
        self.last_token_seen = msg.token
        req = session.get_request(msg.token)
        req.response = True
        session.request_done(req)
    # Saturnin NODE API for clients
    def get_installed(self, **kwargs) -> pb.ReplyInstalledServices:
        """Get list of services installed on Saturnin node.

Returns:
    List of services installed on node.
"""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.INSTALLED_SERVICES, token)
        session.note_request(msg)
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to INSTALLED_SERVICES request")
        return msg.response
    def get_running(self, **kwargs) -> pb.ReplyRunningServices:
        """Get list of services running on Saturnin node.

Returns:
    List of services running on node.
"""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.RUNNING_SERVICES, token)
        session.note_request(msg)
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to RUNNING_SERVICES request")
        return msg.response
    def get_providers(self, interface_uid: UUID, **kwargs) -> t.List[UUID]:
        """Get list of services that provider specified interface.

Returns:
    List of service UIDs.
"""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.INTERFACE_PROVIDERS, token)
        session.note_request(msg)
        dframe = pb.RequestInterfaceProviders()
        dframe.interface_uid = interface_uid.bytes
        msg.data.append(dframe.SerializeToString())
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to INTERFACE_PROVIDERS request")
        return msg.response
    def start_service(self, agent_uid: UUID, config: Config,
                      name: str = None,
                      start_timeout: int = None,
                      singleton: bool = False, **kwargs) -> pb.ReplyStartService:
        """Starts specified service.

Returns:
    Details about started service.
"""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.START_SERVICE, token)
        session.note_request(msg)
        dframe: pb.RequestStartService = pb.RequestStartService()
        dframe.agent_uid = agent_uid.bytes
        config.save_proto(dframe.config)
        if name:
            dframe.name = name
        if start_timeout:
            dframe.timeout = start_timeout
        dframe.singleton = singleton
        msg.data.append(dframe.SerializeToString())
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to START_SERVICE request")
        return msg.response
    def stop_service(self, peer_uid: UUID, stop_timeout: int = None, forced: bool = False,
                      **kwargs) -> pb.ReplyStopService:
        """Starts specified service.

Returns:
    Details about started service.
"""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.STOP_SERVICE, token)
        session.note_request(msg)
        dframe = pb.RequestStopService()
        dframe.peer_uid = peer_uid.bytes
        if stop_timeout:
            dframe.timeout = stop_timeout
        dframe.forced = forced
        msg.data.append(dframe.SerializeToString())
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to STOP_SERVICE request")
        return msg.response
    def shutdown(self, **kwargs) -> True:
        """Stops the Saturnin node."""
        session: Session = self.get_session()
        assert session
        token = self.new_token()
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.SHUTDOWN, token)
        session.note_request(msg)
        self.send(msg)
        if not self.get_response(token, kwargs.get('timeout')):
            raise TimeoutError("The service did not respond on time to SHUTDOWN request")
        return msg.response



#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Runtime node service
# FILE:           saturni.sarvice.node.console.py
# DESCRIPTION:    Saturnin node console (CLI version)
# CREATED:        14.6.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - Runtime node console (CLI version)


"""

import typing as t
import logging
from logging.config import fileConfig
import sys
import os
from uuid import UUID, uuid1
from socket import getfqdn
import cmd
import zmq
from argparse import ArgumentParser, Action, Namespace, FileType
from configparser import ConfigParser, ExtendedInterpolation, DEFAULTSECT
from pkg_resources import iter_entry_points
from tabulate import tabulate
from saturnin.sdk import VENDOR_UID
from saturnin.sdk.types import State, ZMQAddress, AddressDomain, PeerDescriptor, \
     AgentDescriptor, ServiceDescriptor, SaturninError, StopError
from saturnin.sdk.collections import Registry
from saturnin.sdk.config import Config, ServiceConfig, UUIDOption, BoolOption
from saturnin.sdk.base import ChannelManager, DealerChannel
from . import node_pb2 as pb
from .client import SaturninNodeClient

__VERSION__ = '0.1'

CONSOLE_AGENT_UID = UUID('0a82fd0e-9339-11e9-8c3d-5404a6a1fd6e')

SECTION_LOCAL_ADDRESS = 'local_address'
SECTION_NODE_ADDRESS = 'node_address'
SECTION_NET_ADDRESS = 'net_address'
SECTION_SERVICE_UID = 'service_uid'
SECTION_PEER_UID = 'peer_uid'

def title(text: str, size: int = 80, char: str = '='):
    "Returns centered title surrounded by char."
    return f"  {text}  ".center(size, char)

# Classes

class UpperAction(Action):
    "Converts argument to uppercase."
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values.upper())

class NodeConsole(cmd.Cmd):
    """Saturnin - Runtime node console

Attributes:
    :options: Command-line options (arguments)
    :ctx:     ZMQ Context
    :mngr:    ChannelManager
    :chn:     DealerChannel used for communication with node service.
    :peer:    Console PeerDescriptor
    :agent:   Console AgentDescriptor
    :node:    SaturninNodeClient
    :file:    Output file
"""
    def __init__(self, options: Namespace, completekey='tab', stdin=None, stdout=None):
        ""
        super().__init__(completekey, stdin, stdout)
        peer_uid: UUID = uuid1()
        self.options: Namespace = options
        self.ctx: zmq.Context = zmq.Context.instance()
        self.mngr: ChannelManager = ChannelManager(self.ctx)
        self.chn: DealerChannel = DealerChannel(b'node-console:' + peer_uid.hex.encode('ascii'), False)
        self.mngr.add(self.chn)
        self.peer: PeerDescriptor = PeerDescriptor(peer_uid, os.getpid(), getfqdn())
        self.agent: AgentDescriptor = AgentDescriptor(CONSOLE_AGENT_UID,
                                                      "Saturnin node console",
                                                      __VERSION__,
                                                      VENDOR_UID,
                                                      'system/console')
        self.node: SaturninNodeClient = SaturninNodeClient(self.chn, self.peer, self.agent)
        self.file: t.IO = None
        self.log: logging.Logger = logging.getLogger('nodeconsole')
        self.conf: ConfigParser = ConfigParser(interpolation=ExtendedInterpolation())
        self.ext_cfg: Config = Config('console_cfg', "Node console configuration")

        self.opt_svc_uid: UUIDOption = \
            self.ext_cfg.add_option(UUIDOption('service_uid',
                                                   "Service UID (agent.uid in the Service Descriptor)",
                                                   required=True))

        self.opt_svc_single: BoolOption = \
            self.ext_cfg.add_option(BoolOption('singleton',
                                               "When True, start only one service instance",
                                               default=False))
        self.service_registry: Registry = Registry()
        #
        self.__show_completion: t.List = ['installed', 'running', 'providers']
        self.__seen_agents: t.Set = set()
        self.__seen_peers: t.Set = set()
        self.__user_sections: t.List = []
        #
        if 'endpoint' in vars(options) and options.endpoint:
            self.cmdqueue.append('attach %s' % options.endpoint)
    def verbose(self, *args, **kwargs) -> None:
        "Log verbose output, not propagated to upper loggers."
        if self.options.verbose:
            self.log.debug(*args, **kwargs)
    def initialize(self) -> None:
        "Console initialization"
        self.conf[SECTION_LOCAL_ADDRESS] = {}
        self.conf[SECTION_NODE_ADDRESS] = {}
        self.conf[SECTION_NET_ADDRESS] = {}
        self.conf[SECTION_SERVICE_UID] = {}
        self.conf[SECTION_PEER_UID] = {}
        self.conf.read_file(self.options.config)
        # Defaults
        self.conf[DEFAULTSECT]['here'] = os.getcwd()
        if self.options.output_dir is None:
            self.conf[DEFAULTSECT]['output_dir'] = os.getcwd()
        else:
            self.conf[DEFAULTSECT]['output_dir'] = self.options.output_dir
        #
        for section in self.conf.sections():
            if self.conf.has_option(section, 'services'):
                self.__user_sections.append(section.lstrip('run_'))
            elif self.conf.has_option(section, 'service_uid'):
                self.__user_sections.append(section)
        # Logging configuration
        if self.conf.has_section('loggers'):
            self.options.config.seek(0)
            fileConfig(self.options.config)
        else:
            logging.basicConfig(format='%(asctime)s %(processName)s:'\
                                '%(threadName)s:%(name)s %(levelname)s: %(message)s')
        logging.getLogger().setLevel(self.options.log_level)
        # Script output configuration
        self.log.setLevel(logging.DEBUG)
        self.log.propagate = False
        if not self.options.log_only:
            output: logging.StreamHandler = logging.StreamHandler(sys.stdout)
            output.setFormatter(logging.Formatter())
            lvl = logging.INFO
            if self.options.verbose:
                lvl = logging.DEBUG
            elif self.options.quiet:
                lvl = logging.ERROR
            output.setLevel(lvl)
            self.log.addHandler(output)
        self.options.config.close()
        # Installed services
        self.service_registry.extend(entry.load() for entry in
                                     iter_entry_points('saturnin.service'))
        self.conf[SECTION_SERVICE_UID] = dict((sd.agent.name, sd.agent.uid.hex) for sd
                                              in self.service_registry)
        for svc in self.service_registry:
            self.note_agent(svc.agent)
    def finalize(self) -> None:
        "Finalize the console."
        self.close()
        self.node.close()
        self.mngr.shutdown()
        self.ctx.term()
    def close(self):
        "Close the output file if needed."
        if self.file:
            self.file.close()
            self.file = None
    def get_peer_uid(self, svc_name: str) -> UUID:
        """Returns peer uid of running service or None."""
        reply: pb.ReplyRunningServices = self.node.get_running()
        for service in reply.services:
            self.note_peer(service.peer)
            self.note_agent(service.agent)
            if service.agent.name == svc_name:
                return UUID(bytes=service.peer.uid)
        return None
    def note_peer(self, peer):
        "Remembers peer uid"
        self.__seen_peers.add(peer.uid if isinstance(peer.uid, UUID) else UUID(bytes=peer.uid))
    def note_agent(self, agent):
        "Remembers agent uids"
        self.__seen_agents.add(agent.uid if isinstance(agent.uid, UUID) else UUID(bytes=agent.uid))
    def precmd(self, line: str) -> str:
        "Hook method. Converts input to lowercase."
        line = line.lower()
        if self.file and not (line.startswith('run') or line.startswith('eof')):
            print(line, file=self.file)
        return line
    def emptyline(self):
        "We ignore empty lines."
        return
    def do_attach(self, line: str) -> bool:
        """Attach to Saturnin node.

Format:
    ATTACH <endpoint>
"""
        try:
            if self.node.is_connected():
                self.node.close()
                self.chn.drop_socket()
                self.chn.create_socket()
            self.node.open(line)
        except Exception as exc:
            self.log.error('ERROR:', exc)
        else:
            self.log.info("Attached to saturnin node at %s" % line)
    def do_show(self, line: str) -> bool:
        """Show various information about node and services.

Format:
    SHOW INSTALLED - Show information about services installed on node.
    SHOW RUNNING - Show information about services running on node.
    SHOW PROVIDERS <intreface_uid> - Show services that provide specified interface.
"""
        if not self.node.is_connected():
            print("Satunin node not attached.")
            return
        try:
            if line == 'installed':
                reply = self.node.get_installed()
                table = []
                for service in reply.services:
                    self.note_agent(service.agent)
                    table.append([service.agent.name, service.agent.version,
                                  UUID(bytes=service.agent.uid),
                                  service.agent.classification])
                self.log.info("Services installed on node:")
                self.log.info(tabulate(table, headers=('name', 'version', 'agent uid',
                                                       'classification'),
                               tablefmt="grid"))
            elif line == 'running':
                reply: pb.ReplyRunningServices = self.node.get_running()
                table = []
                for service in reply.services:
                    self.note_agent(service.agent)
                    self.note_peer(service.peer)
                    table.append([UUID(bytes=service.peer.uid), service.peer.host,
                                  service.peer.pid, pb.ExecutionModeEnum.Name(service.mode),
                                  service.agent.name, service.agent.version,
                                  '\n'.join(service.endpoints)])
                self.log.info("Services running on node:")
                self.log.info(tabulate(table, headers=('peer uid', 'host', 'pid', 'mode',
                                                       'name', 'version', 'endpoints'),
                               tablefmt="grid"))
            elif line == 'providers':
                pass
            else:
                self.log.error("Unknown SHOW command: %s" % line)
        except Exception as exc:
            self.log.error('ERROR:', exc)
    def complete_show(self, text: str, line: str, begidx: int, endidx: int) -> t.List:
        """Command completion for SHOW command."""
        return [f for f in self.__show_completion if f.startswith(text)]
    def do_start(self, line: str) -> bool:
        """Start service.

Format:
    START <config>[, <config>]

Arguments:
    config: Section name in current configuration file with service specification
"""
        if not self.node.is_connected():
            print("Satunin node not attached.")
            return
        try:
            # Create list of service sections
            sections = []
            for job_name in (x.strip() for x in line.split(', ')):
                job_section = 'run_%s' % job_name
                if self.conf.has_section(job_name):
                    sections.append(job_name)
                elif self.conf.has_section(job_section):
                    if not self.conf.has_option(job_section, 'services'):
                        raise StopError("Missing 'services' option in section '%s'" %
                                        job_section)
                    for name in (value.strip() for value in self.conf.get(job_section,
                                                                          'services').split(',')):
                        if not self.conf.has_section(name):
                            raise StopError("Configuration does not have section '%s'" % name)
                        sections.append(name)
                else:
                    raise StopError("Configuration does not have section '%s' or '%s'" %
                                    (job_name, job_section))
            # Validate configuration of services
            services = []
            for svc_section in sections:
                if not self.conf.has_option(svc_section, self.opt_svc_uid.name):
                    raise StopError("Missing '%s' option in section '%s'" % (self.opt_svc_uid.name,
                                                                             svc_section))
                self.ext_cfg.clear()
                self.ext_cfg.load_from(self.conf, svc_section)
                svc: ServiceDescriptor = self.service_registry.get(self.opt_svc_uid.value)
                if svc is None:
                    raise StopError("Unknown service '%s'" % self.opt_svc_uid.value)
                cfg: ServiceConfig = svc.config()
                try:
                    cfg.load_from(self.conf, svc_section)
                    cfg.validate()
                except (SaturninError, TypeError, ValueError) as exc:
                    raise StopError("Error in configuration section '%s'\n%s" % \
                                    (svc_section, str(exc)))
                services.append((svc_section, svc, cfg, self.opt_svc_single.value))
            #
            for section_name, svc, cfg, singleton in services:
                # print configuration
                svc: ServiceDescriptor
                cfg: ServiceConfig
                self.verbose(title("Task '%s'" % section_name, char='-'))
                self.verbose("service_uid = %s [%s]" % (svc.agent.uid, svc.agent.name))
                self.verbose("singleton   = %s" % ('Yes' if singleton else 'No'))
                for option in cfg.options.values():
                    self.verbose("%s" % option.get_printout())
            self.verbose(title("Starting services"))
            #
            for section_name, svc, cfg, singleton in services:
                # refresh configuration to fetch actual addresses
                self.log.info("Starting service '%s', task '%s'", svc.agent.name,
                              section_name)
                cfg.load_from(self.conf, section_name)
                #
                reply: pb.ReplyStartService = \
                    self.node.start_service(svc.agent.uid, cfg, section_name,
                                            singleton=singleton)
                #
                if reply is None:
                    print('start_service returned NONE, no exception')
                    return
                if reply.service.endpoints:
                    # Update addresses
                    for endpoint in reply.service.endpoints:
                        endpoint = ZMQAddress(endpoint)
                        if endpoint.domain == AddressDomain.LOCAL:
                            self.conf[SECTION_LOCAL_ADDRESS][section_name] = endpoint
                        elif endpoint.domain == AddressDomain.NODE:
                            self.conf[SECTION_NODE_ADDRESS][section_name] = endpoint
                        else:
                            self.conf[SECTION_NET_ADDRESS][section_name] = endpoint
                #
                table = []
                table.append(['peer uid', UUID(bytes=reply.service.peer.uid)])
                table.append(['mode', pb.ExecutionModeEnum.Name(reply.service.mode)])
                table.append(['endpoints', '\n'.join(reply.service.endpoints)])
                self.log.info("Service started:")
                self.log.info(tabulate(table, tablefmt="grid"))
        except SaturninError as exc:
            self.log.error('ERROR: %s', exc)
        except Exception as exc:
            self.log.exception(exc)
    def complete_start(self, text: str, line: str, begidx: int, endidx: int) -> t.List:
        """Command completion for START command."""
        return [x for x in self.__user_sections if x.startswith(text)]
    def do_stop(self, line: str) -> bool:
        """Stop running service.

Format:
    STOP <service_uid> | ALL
"""
        if not self.node.is_connected():
            self.log.error("Satunin node not attached.")
            return
        uids = []
        if line.strip().upper() == 'ALL':
            reply: pb.ReplyRunningServices = self.node.get_running()
            for service in reply.services:
                uids.append(UUID(bytes=service.peer.uid))
        else:
            try:
                uid = UUID(line)
            except ValueError:
                uid = self.get_peer_uid(line)
                if not uid:
                    self.log.error("Service '%s' is not running." % line)
                    return
            uids.append(uid)
        for uid in uids:
            try:
                reply = self.node.stop_service(uid)
                state = 'STOPPED' if State(reply) == State.STOPPED else State(reply).name
                self.log.info("Service %s %s" % (uid, state))
            except Exception as exc:
                self.log.error('ERROR:', exc)
    def complete_stop(self, text: str, line: str, begidx: int, endidx: int) -> t.List:
        """Command completion for STOP command."""
        return [x for x in (str(x) for x in self.__seen_peers) if x.startswith(text)]
    def do_shutdown(self, line: str) -> bool:
        """Shuts down the Saturnin node, terminating all running services.
Format:
    SHUTDOWN
"""
        if not self.node.is_connected():
            print("Satunin node not attached.")
            return
        try:
            self.node.shutdown()
            self.node.close()
            self.chn.drop_socket()
            self.chn.create_socket()
            print("Saturnin node successfully shut down.")
        except Exception as exc:
            print('ERROR:', exc)
    def do_exit(self, line: str) -> bool:
        """Terminate the node console.

Format:
    EXIT
"""
        return True
    def do_record(self, arg):
        """Save future commands to file.

Format:
    RECORD <filename>
"""
        self.file = open(arg, 'w')
    def do_eof(self, arg):
        """Stop recording commands started by RECORD, and close the file.

Format:
    EOF
"""
        self.close()
    def do_run(self, arg):
        """Playback commands from a file.

Format:
    RUN <filename>
"""
        self.close()
        with open(arg) as f:
            self.cmdqueue.extend(f.read().splitlines())

def main():
    "Main function"

    into = """Saturnin node console v%s
Type 'help' or '?' for list of commands.
""" % __VERSION__

    logging.basicConfig(format='%(levelname)s:%(threadName)s:%(name)s:%(message)s')
    description = """Saturnin node console."""
    parser = ArgumentParser(description=description)
    parser.add_argument('--version', action='version', version='%(prog)s '+__VERSION__)

    group = parser.add_argument_group("run arguments")
    group.add_argument('-c', '--config', metavar='FILE',
                       type=FileType(mode='r', encoding='utf8'),
                       help="Configuration file")
    group.add_argument('-e', '--endpoint', help="ZMQ addresses of the Saturnin node")
    group.add_argument('-o', '--output-dir', metavar='DIR',
                       help="Force directory for log files and other output")

    group = parser.add_argument_group("output arguments")
    group.add_argument('-v', '--verbose', action='store_true', help="Verbose output")
    group.add_argument('-q', '--quiet', action='store_true', help="No screen output")
    group.add_argument('--log-only', action='store_true',
                       help="Suppress all screen output including error messages")
    group.add_argument('-l', '--log-level', action=UpperAction,
                       choices=[x.lower() for x in logging._nameToLevel
                                if isinstance(x, str)],
                       help="Logging level")
    group.add_argument('--trace', action='store_true',
                       help="Log unexpected errors with stack trace")
    parser.set_defaults(log_level='WARNING', config='nodeconsole.cfg',
                             output_dir='${here}')

    args = parser.parse_args()
    logger = logging.getLogger()
    logger.setLevel(args.log_level)

    console = NodeConsole(args)
    try:
        console.initialize()
        console.cmdloop(into)
    except KeyboardInterrupt:
        pass
    finally:
        console.finalize()




#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Runtime node service
# FILE:           saturnin/service/node/test.py
# DESCRIPTION:    Test runner for NODE service (classic version)
# CREATED:        13.5.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Test runner for NODE service (classic version)
"""

from uuid import UUID
import zmq
from saturnin.sdk.types import ZMQAddress
from saturnin.sdk.base import Channel
from saturnin.sdk.test.fbsp import MsgType, BaseTestRunner, print_msg, print_title, \
     print_data
from saturnin.sdk.config import ServiceConfig
from . import node_pb2 as pb
from .api import NodeRequest, NODE_INTERFACE_UID
from .client import SaturninNodeClient
import saturnin.service.roman.api as roman_api

class TestRunner(BaseTestRunner):
    """Test Runner for NODE Service and Client.
"""
    def __init__(self, context):
        super().__init__(context)
        self.interface_uid = NODE_INTERFACE_UID.bytes
        self.peer_uid = None
    def create_client(self, channel: Channel) -> SaturninNodeClient:
        return SaturninNodeClient(channel, self.peer, self.agent)
    def run_request(self, api_call, *args, **kwargs):
        "Execute Client API call."
        print('Sending request..')
        try:
            data = api_call(*args, **kwargs)
            print('Received:')
            print_data(data)
            return data
        except Exception as exc:
            print_title("ERROR", char="*")
            print(exc)
    def raw_01_installed_services(self, socket: zmq.Socket):
        "Raw test of INSTALLED_SERVICES request."
        print("Sending INSTALLED_SERVICES request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.INSTALLED_SERVICES,
                                               self.get_token())
        print_msg(msg)
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        dframe = pb.ReplyInstalledServices()
        dframe.ParseFromString(msg.data[0])
        print_msg(msg, str(dframe))
    def raw_02_interface_providers(self, socket: zmq.Socket):
        "Raw test of INTERFACE_PROVIDERS request."
        print("Sending INTERFACE_PROVIDERS request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.INTERFACE_PROVIDERS,
                                               self.get_token())
        dframe = pb.RequestInterfaceProviders()
        dframe.interface_uid = roman_api.ROMAN_INTERFACE_UID.bytes
        msg.data.append(dframe.SerializeToString())
        print_msg(msg, str(dframe))
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        dframe = pb.ReplyInterfaceProviders()
        dframe.ParseFromString(msg.data[0])
        print_msg(msg, str(dframe))
        for agent_uid in dframe.agent_uids:
            if UUID(bytes=agent_uid) == roman_api.SERVICE_UID:
                print("    The reply is correct")
                return
    def raw_03_start_service(self, socket: zmq.Socket):
        "Raw test of START_SERVICE request."
        print("Sending START_SERVICE request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.START_SERVICE,
                                               self.get_token())
        dframe: pb.RequestStartService = pb.RequestStartService()
        dframe.agent_uid = roman_api.SERVICE_UID.bytes
        cfg: ServiceConfig = roman_api.SERVICE_DESCRIPTION.config()
        cfg.endpoints.set_value([ZMQAddress('tcp://127.0.0.1:7001')])
        cfg.save_proto(dframe.config)
        dframe.name = 'roman-test'
        msg.data.append(dframe.SerializeToString())
        print_msg(msg, str(dframe))
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        if msg.message_type == MsgType.REPLY:
            dframe = pb.ReplyStartService()
            dframe.ParseFromString(msg.data[0])
            self.peer_uid = dframe.service.peer.uid
            data = str(dframe)
        elif msg.message_type == MsgType.ERROR:
            data = None
        print_msg(msg, data)
    def raw_04_running_services(self, socket: zmq.Socket):
        "Raw test of RUNNING_SERVICES request."
        print("Sending RUNNING_SERVICES request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.RUNNING_SERVICES,
                                               self.get_token())
        print_msg(msg)
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        dframe = pb.ReplyRunningServices()
        dframe.ParseFromString(msg.data[0])
        print_msg(msg, str(dframe))
    def raw_05_stop_service(self, socket: zmq.Socket):
        "Raw test of STOP_SERVICE request."
        if not self.peer_uid:
            print_title("TEST SKIPPED", char='*')
            return
        print("Sending STOP_SERVICE request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.STOP_SERVICE,
                                               self.get_token())
        dframe = pb.RequestStopService()
        dframe.peer_uid = self.peer_uid
        msg.data.append(dframe.SerializeToString())
        print_msg(msg, str(dframe))
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        if msg.message_type == MsgType.REPLY:
            dframe = pb.ReplyStopService()
            dframe.ParseFromString(msg.data[0])
            data = str(dframe)
        elif msg.message_type == MsgType.ERROR:
            data = None
        print_msg(msg, data)
    def raw_06_shutdown(self, socket: zmq.Socket):
        "Raw test of SHUTDOWN request."
        print("Sending SHUTDOWN request:")
        msg = self.protocol.create_request_for(self.interface_id,
                                               NodeRequest.SHUTDOWN,
                                               self.get_token())
        print_msg(msg, None)
        socket.send_multipart(msg.as_zmsg())
        print("Receiving reply:")
        zmsg = socket.recv_multipart()
        msg = self.protocol.parse(zmsg)
        print_msg(msg, None)
    def client_01_installed_services(self, client: SaturninNodeClient):
        "Client test of get_installed() API call."
        self.run_request(client.get_installed)
    def client_02_interface_providers(self, client: SaturninNodeClient):
        "Client test of get_providers() API call."
        self.run_request(client.get_providers, roman_api.ROMAN_INTERFACE_UID)
    def client_03_start_service(self, client: SaturninNodeClient):
        "Client test of start_service() API call."
        cfg: ServiceConfig = roman_api.SERVICE_DESCRIPTION.config()
        cfg.endpoints.set_value([ZMQAddress('tcp://127.0.0.1:*')])
        data = self.run_request(client.start_service, roman_api.SERVICE_UID, cfg)
        self.peer_uid = UUID(bytes=data.service.peer.uid)
    def client_04_running_services(self, client: SaturninNodeClient):
        "Client test of get_running() API call."
        self.run_request(client.get_running)
    def client_05_stop_service(self, client: SaturninNodeClient):
        "Client test of stop_service() API call."
        self.run_request(client.stop_service, self.peer_uid)
    def client_06_shutdown(self, client: SaturninNodeClient):
        "Client test of get_running() API call."
        self.run_request(client.shutdown)



#coding:utf-8
"""A setuptools based setup module for saturnin package.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from codecs import open
from os import path
# Always prefer setuptools over distutils
from setuptools import setup, find_packages

HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='saturnin',
    version='0.2.0',
    description='Reference implementation of selected Firebird Butler services in Python',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url='https://github.com/FirebirdSQL/saturnin',
    author='Pavel Císař',
    author_email='pcisar@users.sourceforge.net',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',

        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',

#        'Framework :: Saturnin SDK',

        'Topic :: Database',
        'Topic :: System :: Systems Administration',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Logging'
        ],
    keywords='Firebird Butler Services ZeroMQ',  # Optional
    packages=['saturnin'],  # Required
    install_requires=['pyzmq>=18.0.0', 'protobuf>=3.9.0', 'saturnin-sdk>=0.4.0',
                      'tabulate>=0.8.3'],
    python_requires='>=3.6, <4',
    test_suite='nose.collector',
    data_files=[],
    namespace_packages=[],
    project_urls={
        'Documentation': 'http://saturnin.readthedocs.io/en/latest/',
        'Bug Reports': 'https://github.com/FirebirdSQL/saturnin/issues',
        'Funding': 'https://www.firebirdsql.org/en/donate/',
        'Source': 'https://github.com/FirebirdSQL/saturnin',
    },
    entry_points={'console_scripts': ['saturnin-node = saturnin.service.node.runner:main',
                                      'nodeconsole = saturnin.service.node.console:main',
                                     ],
                  'saturnin.service': ['saturnin-node = saturnin.service.node.api:SERVICE_DESCRIPTION',
                                       'firebird-log = saturnin.micro.fblog.api:SERVICE_DESCRIPTION',
                                      ],
                 }
)



#coding:utf-8
#
# PROGRAM/MODULE: Saturnin - Runtime node service
# FILE:           saturnin/service/node/service.py
# DESCRIPTION:    Saturnin runtime node service (classic version)
# CREATED:        12.5.2019
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2019 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________.

"""Saturnin - Runtime node service (classic version)

NODE Service manages Saturnin runtime node. It provides environment for execution and
management of other Saturnin services.

Supported requests:

    :INSTALLED_SERVICES:  REPLY with list of installed services available for execution on the node.
    :RUNNING_SERVICES:    REPLY with list of services actually running on the node.
    :INTERFACE_PROVIDERS: REPLY with list of services that provide specified interface.
    :START_SERVICE:       Start service on node. REPLY with service instance information.
    :STOP_SERVICE:        Stop service running on node. REPLY with stop state.
    :SHUTDOWN:            Shuts down the NODE service.
"""

import logging
import typing as t
from uuid import UUID
from pkg_resources import iter_entry_points
from firebird.butler.fbsd_pb2 import STATE_STOPPED, STATE_TERMINATED, STATE_UNKNOWN, \
     STATE_RUNNING
from saturnin.sdk.types import ExecutionMode, DependencyType, AddressDomain, ZMQAddress, \
     ZMQAddressList, ServiceFacilities, ServiceDescriptor, ServiceError, SaturninError
from saturnin.sdk.collections import Registry
from saturnin.sdk.protocol.fbsp import fbsp_proto, MsgType, ErrorCode, Session, \
     ServiceMessagelHandler, HelloMessage, CancelMessage, RequestMessage, bb2h
from saturnin.sdk.service import TEvent, BaseService, SimpleServiceImpl
from saturnin.sdk.classic import ServiceExecutor
from . import node_pb2 as pb
from .api import NodeRequest, NodeError

# Logger

log = logging.getLogger(__name__)

# Constants

DEFAULT_TIMEOUT = 10000

DescriptorList = t.List[ServiceDescriptor]

# Functions

def get_best_endpoint(endpoints: ZMQAddressList,
                      client_mode: ExecutionMode = ExecutionMode.PROCESS,
                      service_mode: ExecutionMode = ExecutionMode.PROCESS) -> t.Optional[ZMQAddress]:
    """Returns endpoint address that uses the best protocol from available options, or None
if there is no address that could be used.
"""
    if endpoints is None:
        return None
    if client_mode == ExecutionMode.ANY:
        client_mode = ExecutionMode.THREAD
    if service_mode == ExecutionMode.ANY:
        service_mode = ExecutionMode.THREAD
    inproc = [x for x in endpoints if x.domain == AddressDomain.LOCAL]
    if (inproc and client_mode == ExecutionMode.THREAD and service_mode == ExecutionMode.THREAD):
        return inproc[0]
    ipc = [x for x in endpoints if x.domain == AddressDomain.NODE]
    if ipc:
        return ipc[0]
    tcp = [x for x in endpoints if x.domain == AddressDomain.NETWORK]
    return tcp[0] if tcp else None

# Classes

class SaturninNodeMessageHandler(ServiceMessagelHandler):
    """Message handler for Saturnin NODE service."""
    RQ_INSTALLED_SERVICES = bb2h(1, NodeRequest.INSTALLED_SERVICES)
    RQ_RUNNING_SERVICES = bb2h(1, NodeRequest.RUNNING_SERVICES)
    RQ_INTERFACE_PROVIDERS = bb2h(1, NodeRequest.INTERFACE_PROVIDERS)
    RQ_START_SERVICE = bb2h(1, NodeRequest.START_SERVICE)
    RQ_STOP_SERVICE = bb2h(1, NodeRequest.STOP_SERVICE)
    RQ_SHUTDOWN = bb2h(1, NodeRequest.SHUTDOWN)
    def __init__(self, welcome_df: fbsp_proto.FBSPWelcomeDataframe, impl: SimpleServiceImpl):
        super().__init__()
        self.impl: SaturninNodeServiceImpl = impl
        self.welcome_df = welcome_df
        # Our message handlers
        self.handlers.update({(MsgType.REQUEST, self.RQ_INSTALLED_SERVICES): self.handle_installed,
                              (MsgType.REQUEST, self.RQ_RUNNING_SERVICES): self.handle_running,
                              (MsgType.REQUEST, self.RQ_INTERFACE_PROVIDERS): self.handle_providers,
                              (MsgType.REQUEST, self.RQ_START_SERVICE): self.handle_start,
                              (MsgType.REQUEST, self.RQ_STOP_SERVICE): self.handle_stop,
                              (MsgType.REQUEST, self.RQ_SHUTDOWN): self.handle_shutdown,
                              MsgType.DATA: self.send_protocol_violation,
                             })
    def set_running_info(self, run_info: pb.RunningService, svc: ServiceExecutor) -> None:
        "Fill RunningService protobuf from ServiceExecutor information."
        # Peer
        run_info.peer.uid = svc.peer.uid.bytes
        run_info.peer.pid = svc.peer.pid
        run_info.peer.host = svc.peer.host
        # Agent
        run_info.agent.uid = svc.descriptor.agent.uid.bytes
        run_info.agent.name = svc.descriptor.agent.name
        run_info.agent.version = svc.descriptor.agent.version
        run_info.agent.vendor.uid = svc.descriptor.agent.vendor_uid.bytes
        run_info.agent.platform.uid = svc.descriptor.agent.platform_uid.bytes
        run_info.agent.platform.version = svc.descriptor.agent.platform_version
        run_info.agent.classification = svc.descriptor.agent.classification
        # Rest
        run_info.description = svc.descriptor.description
        run_info.mode = svc.mode.value
        for flag in svc.facilities.get_flags():
            if flag == ServiceFacilities.FBSP_SOCKET:
                run_info.facilities.append(pb.SVC_FACILITY_FBSP_SOCKET)
            elif flag == ServiceFacilities.INPUT_SERVER:
                run_info.facilities.append(pb.SVC_FACILITY_INPUT_SERVER)
            elif flag == ServiceFacilities.INPUT_CLIENT:
                run_info.facilities.append(pb.SVC_FACILITY_INPUT_CLIENT)
            elif flag == ServiceFacilities.OUTPUT_SERVER:
                run_info.facilities.append(pb.SVC_FACILITY_OUTPUT_SERVER)
            elif flag == ServiceFacilities.OUTPUT_CLIENT:
                run_info.facilities.append(pb.SVC_FACILITY_OUTPUT_CLIENT)
        svc.config.save_proto(run_info.config)
        if svc.endpoints:
            run_info.endpoints.extend(svc.endpoints)
    def handle_hello(self, session: Session, msg: HelloMessage):
        "HELLO message handler. Sends WELCOME message back to the client."
        if __debug__:
            log.debug("%s.handle_hello(%s)", self.__class__.__name__, session.routing_id)
        super().handle_hello(session, msg)
        welcome = self.protocol.create_welcome_reply(msg)
        welcome.peer.CopyFrom(self.welcome_df)
        self.send(welcome, session)
    def handle_cancel(self, session: Session, msg: CancelMessage):
        "Handle CANCEL message."
        # NODE uses simple REQUEST/REPLY API, so there is no point to send CANCEL
        # messages. However, we have to handle it although we'll do nothing.
        # In such cases we could either override the on_cancel() method like now,
        # or assign self.do_nothing handler to MsgType.CANCEL in __init__().
        log.debug("%s.on_cancel(%s)", self.__class__.__name__, session.routing_id)
    def handle_installed(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/INSTALLED_SERVICES message.
"""
        log.debug("%s.on_installed(%s)", self.__class__.__name__, session.routing_id)
        reply = self.protocol.create_reply_for(msg)
        # create reply data frame
        dframe = pb.ReplyInstalledServices()
        for svc_desc in (svc for svc in self.impl.installed_services):
            svc_data = dframe.services.add()
            # Agent
            svc_data.agent.uid = svc_desc.agent.uid.bytes
            svc_data.agent.name = svc_desc.agent.name
            svc_data.agent.version = svc_desc.agent.version
            svc_data.agent.vendor.uid = svc_desc.agent.vendor_uid.bytes
            svc_data.agent.platform.uid = svc_desc.agent.platform_uid.bytes
            svc_data.agent.platform.version = svc_desc.agent.platform_version
            svc_data.agent.classification = svc_desc.agent.classification
            # API
            for api in svc_desc.api:
                api_pb = svc_data.api.add()
                api_pb.uid = api.uid.bytes
                api_pb.name = api.name
                api_pb.revision = api.revision
                api_pb.number = api.number
                # Request codes
                for rcode in api.requests:
                    rc_pb = api_pb.requests.add()
                    rc_pb.code = rcode.value
                    rc_pb.name = rcode.name
            # Dependencies
            for d_type, d_uid in svc_desc.dependencies:
                d_pb = svc_data.dependencies.add()
                d_pb.type = d_type.value
                d_pb.uid = d_uid.bytes
            #
            svc_data.service_type = svc_desc.service_type
            svc_data.description = svc_desc.description
            svc_data.mode = svc_desc.execution_mode
            for flag in svc_desc.facilities.get_flags():
                if flag == ServiceFacilities.FBSP_SOCKET:
                    svc_data.facilities.append(pb.SVC_FACILITY_FBSP_SOCKET)
                elif flag == ServiceFacilities.INPUT_SERVER:
                    svc_data.facilities.append(pb.SVC_FACILITY_INPUT_SERVER)
                elif flag == ServiceFacilities.INPUT_CLIENT:
                    svc_data.facilities.append(pb.SVC_FACILITY_INPUT_CLIENT)
                elif flag == ServiceFacilities.OUTPUT_SERVER:
                    svc_data.facilities.append(pb.SVC_FACILITY_OUTPUT_SERVER)
                elif flag == ServiceFacilities.OUTPUT_CLIENT:
                    svc_data.facilities.append(pb.SVC_FACILITY_OUTPUT_CLIENT)
        reply.data.append(dframe.SerializeToString())
        self.send(reply, session)
    def handle_running(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/RUNNING_SERVICES message.
"""
        log.debug("%s.on_running(%s)", self.__class__.__name__, session.routing_id)
        reply = self.protocol.create_reply_for(msg)
        dframe = pb.ReplyRunningServices()
        for svc in self.impl.services:
            svc: ServiceExecutor
            if svc.is_running():
                self.set_running_info(dframe.services.add(), svc)
        reply.data.append(dframe.SerializeToString())
        self.send(reply, session)
    def handle_providers(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/INTERFACE_PROVIDERS message.
"""
        log.debug("%s.on_providers(%s)", self.__class__.__name__, session.routing_id)
        dframe = pb.RequestInterfaceProviders()
        dframe.ParseFromString(msg.data[0])
        interface_uid = dframe.interface_uid
        # create reply
        reply = self.protocol.create_reply_for(msg)
        dframe = pb.ReplyInterfaceProviders()
        for svc_desc in self.impl.get_providers(interface_uid):
            dframe.agent_uids.append(svc_desc.agent.uid.bytes)
        reply.data.append(dframe.SerializeToString())
        self.send(reply, session)
    def handle_start(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/START_SERVICE message.
"""
        if __debug__: log.debug("%s.on_start(%s)", self.__class__.__name__, session.routing_id)
        dframe: pb.RequestStartService = pb.RequestStartService()
        dframe.ParseFromString(msg.data[0])
        agent_uid = UUID(bytes=dframe.agent_uid)
        agent: ServiceDescriptor = self.impl.installed_services.get(agent_uid)
        if agent is None:
            self.send_error(session, msg, ErrorCode.NOT_FOUND, "Service not found")
            return
        #
        if self.impl.is_running(agent_uid) and dframe.singleton:
            reply = self.protocol.create_reply_for(msg)
            dframe = pb.ReplyStartService()
            svc = self.impl.get_running(agent_uid)
            if svc:
                self.set_running_info(dframe.service, svc)
                reply.data.append(dframe.SerializeToString())
                self.send(reply, session)
                return
        #
        try:
            svc: ServiceExecutor = self.impl.add_service(agent)
        except ServiceError as exc:
            self.send_error(session, msg, ErrorCode.ERROR, "Service start failed",
                            app_code=NodeError.START_FAILED, exc=exc)
            return
        if dframe.name:
            svc.name = dframe.name
        cfg = svc.descriptor.config()
        cfg.load_proto(dframe.config)
        try:
            if dframe.timeout > 0:
                svc.start(cfg, timeout=dframe.timeout)
            else:
                svc.start(cfg)
        except TimeoutError:
            self.send_error(session, msg, ErrorCode.REQUEST_TIMEOUT, "Service start failed")
            return
        except ServiceError as exc:
            self.send_error(session, msg, ErrorCode.ERROR, "Service start failed",
                            app_code=NodeError.START_FAILED, exc=exc)
            return
        # reply
        reply = self.protocol.create_reply_for(msg)
        dframe = pb.ReplyStartService()
        self.set_running_info(dframe.service, svc)
        reply.data.append(dframe.SerializeToString())
        self.send(reply, session)
    def handle_stop(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/STOP_SERVICE message.
"""
        log.debug("%s.on_stop(%s)", self.__class__.__name__, session.routing_id)
        dframe: pb.RequestStopService = pb.RequestStopService()
        dframe.ParseFromString(msg.data[0])
        peer_uid = UUID(bytes=dframe.peer_uid)
        svc: ServiceExecutor = self.impl.services.get(peer_uid)
        if svc is None:
            self.send_error(session, msg, ErrorCode.NOT_FOUND, "Service is not running")
            return
        #
        rframe: pb.ReplyStopService = pb.ReplyStopService()
        rframe.result = STATE_UNKNOWN
        try:
            if dframe.timeout > 0:
                svc.stop(dframe.timeout / 1000)
            else:
                svc.stop()
            rframe.result = STATE_STOPPED
        except TimeoutError:
            if not dframe.forced:
                self.send_error(session, msg, ErrorCode.REQUEST_TIMEOUT,
                                "Failed to stop service",
                                app_code=NodeError.TERMINATION_FAILED)
                return
        except ServiceError as exc:
            if not dframe.forced:
                self.send_error(session, msg, ErrorCode.ERROR, "Failed to stop service",
                                app_code=NodeError.TERMINATION_FAILED, exc=exc)
                return
        if rframe.result == STATE_UNKNOWN:
            try:
                svc.terminate()
                rframe.result = STATE_TERMINATED
            except SaturninError as exc:
                self.send_error(session, msg, ErrorCode.ERROR,
                                "Service termination has failed",
                                app_code=NodeError.TERMINATION_FAILED, exc=exc)
                return
        #
        if svc.is_running():
            rframe.result = STATE_RUNNING
        else:
            self.impl.services.remove(svc)
        #
        reply = self.protocol.create_reply_for(msg)
        reply.data.append(rframe.SerializeToString())
        self.send(reply, session)
    def handle_shutdown(self, session: Session, msg: RequestMessage):
        """Handle REQUEST/SHUTDOWN message.
"""
        log.debug("%s.on_shutdown(%s)", self.__class__.__name__, session.routing_id)
        # send reply to confirm that shutdown was initiated
        reply = self.protocol.create_reply_for(msg)
        self.send(reply, session)
        # commence shutdown
        self.impl.stop_event.set()

class SaturninNodeServiceImpl(SimpleServiceImpl):
    """Implementation of Saturnin NODE service."""
    def __init__(self, descriptor: ServiceDescriptor, stop_event: TEvent):
        super().__init__(descriptor, stop_event)
        self.installed_services: Registry = Registry()
        self.services: Registry = Registry()
    def initialize(self, svc: BaseService):
        super().initialize(svc)
        self.svc_chn.set_handler(SaturninNodeMessageHandler(self.welcome_df, self))
        # Load information about installed services.
        self.installed_services.extend(entry.load() for entry in
                                       iter_entry_points('saturnin.service'))
    def finalize(self, svc: BaseService) -> None:
        """Service finalization. Stops/terminates all services running on node.
"""
        log.debug("%s.finalize", self.__class__.__name__)
        for svc in self.services:
            svc: ServiceExecutor
            try:
                svc.stop()
            except:
                try:
                    log.warning("Service did not stop on time - terminating, peer[%s:%s] agent[%s:%s]",
                                svc.uid, svc.name, svc.agent_uid, svc.agent_name)
                    svc.terminate()
                except:
                    log.exception("Service termination has failed, peer[%s:%s] agent[%s:%s]",
                                  svc.uid, svc.name, svc.agent_uid, svc.agent_name)
        super().finalize(svc)
    def on_idle(self) -> None:
        """Called by service when waiting for messages exceeds timeout. Performs check
for running services. Dead services are removed from list of running services.
"""
        for svc in self.services:
            if not svc.is_running():
                log.info("The service ended itself, peer[%s:%s] agent[%s:%s]",
                         svc.uid, svc.name, svc.agent_uid, svc.agent_name)
                self.services.remove(svc)
    def get_service_descriptor(self, agent_uid: UUID) -> ServiceDescriptor:
        """Returns descriptor for installed service.

Raises:
    :ValueError: If service is not installed.
"""
        return self.installed_services.get(agent_uid)
    def get_providers(self, interface_uid: bytes) -> DescriptorList:
        """Returns list of ServiceDescriptors of services that implement interface."""
        result = []
        for svc_desc in self.installed_services:
            for api in svc_desc.api:
                if api.uid.bytes == interface_uid:
                    result.append(svc_desc)
        return result
    def get_running_providers(self, interface_uid: bytes) -> DescriptorList:
        """Returns list of ServiceDescriptors of services that implement interface."""
        result = []
        for svc in self.services:
            svc: ServiceExecutor
            for api in svc.descriptor.api:
                if api.uid.bytes == interface_uid:
                    result.append(svc.descriptor)
        return result
    def add_service(self, svc_descriptor: ServiceDescriptor) -> ServiceExecutor:
        """Create and return new `ServiceExecutor`.

Raises:
    :ServiceError: If the service has declared the required dependencies that are not running.
"""
        def find_provider_address(interface_uid):
            for provider in self.get_running_providers(interface_uid.bytes):
                provider: ServiceExecutor
                for svc in self.services:
                    if svc.agent_uid == provider.agent_uid:
                        return get_best_endpoint(svc.endpoints, provider.mode, svc.mode)
            return None

        # Find endpoints to running dependencies
        svc = ServiceExecutor(svc_descriptor)
        for dtype, interface_uid in svc_descriptor.dependencies:
            if dtype == DependencyType.REQUIRED:
                if find_provider_address(interface_uid) is None:
                    raise ServiceError("Provider not available for required interface %s"
                                       % interface_uid)
        self.services.store(svc)
        return svc
    def get_running(self, agent_uid: UUID) -> bool:
        """Returns True if service is running."""
        for svc in self.services:
            if svc.agent_uid == agent_uid:
                if svc.is_running():
                    return svc
        return None
    def is_running(self, agent_uid: UUID) -> bool:
        """Returns True if service is running."""
        for svc in self.services:
            if svc.agent_uid == agent_uid:
                if svc.is_running():
                    return True
        return False
